# AUTOGENERATED! DO NOT EDIT! File to edit: 12_synchro.processing.ipynb (unless otherwise specified).

__all__ = ['get_thresholds', 'get_first_high', 'detect_frames', 'reverse_detection', 'extend_timepoints', 'error_check',
           'cluster_frame_signals', 'cluster_by_epochs', 'cluster_by_list', 'parse_time', 'get_position_estimate',
           'match_starting_position', 'display_match', 'frame_error_correction', 'error_frame_matches', 'apply_shifts',
           'shift_detection_conv', 'shift_detection_NW', 'chop_stim_edges', 'detect_calcium_frames']

# Cell
import numpy as np
import datetime
import glob
import os
from scipy import signal

# Cell
def get_thresholds(data):
    """Function that attempts to get the high and low thresholds. Not working very well"""
    max_val = max(data[len(data)//2:len(data)//2 + 10000000]) #Looking for a max in a portion of the data, from the middle
    high_thresh = max_val*3/4 # High threshold set at 3/4th of the max
    low_thresh  = max_val*1/4
    return low_thresh, high_thresh

# Cell
def get_first_high(data, threshold):
    if np.any(data>threshold):
        return np.argmax(data>threshold)
    else:
        return -1

# Cell
def detect_frames(data, low_threshold, high_threshold, increment, do_reverse=True, precision=.95):
    """Frame detection (or ON signal detection). Capable of finding frame times produced in a regular
    fashion:
        - data: raw data
        - low_threshold: threshold used to detect begginning of each frame.
        - high_threshold: threshold used to assign label to the frames, and used to detect the beggining of the reading frame.
        - increment: Number of timepoints separating frames. Eg, for 30KHz recording of 60hz stimulus: 30000/60
        - do_reverse: boolean to indicate if the reverse detection should be done after detecting the first frame.
        - precision: Value that indicates how precise are the events recorded to accelerate the detection.
        DLP is very stable (.95) whereas some camera triggers have more jitter (.6). Too low value (bellow 0.5) can
        result in an overdetection of frames.

    """
    assert (precision>0) and (precision<=1)
    frame_timepoints, frame_signals = [], []
    increment      = int(increment)
    safe_increment = int(increment*precision)

    first_high = get_first_high(data, high_threshold)
    if first_high == -1:
        print("No high frame detected. Detection can't work.")
        return

    frame_timepoints.append(first_high)
    frame_signals.append(1)

    if do_reverse:
        new_timepoints   = reverse_detection(data, frame_timepoints, low_threshold, increment, precision)
        if len(new_timepoints)>1:
            new_extrapolated = extend_timepoints(new_timepoints)
        else:
            new_extrapolated = []
        frame_timepoints = new_extrapolated + new_timepoints + frame_timepoints
        frame_signals    = [0]*(len(new_timepoints)+len(new_extrapolated)) + frame_signals

    i = first_high + safe_increment
    while i < len(data):
        data_slice = data[i:i+increment//2+(increment-safe_increment)*2]
        if np.any(data_slice>low_threshold):
            i = i+np.argmax(data_slice>low_threshold)
        else:
            break #This frame sequence is over. Pass the next sequence through this function if there are frames left
        frame_timepoints.append(i)
        frame_signals.append(int(np.any(data_slice > high_threshold)))
        i += safe_increment

    frame_timepoints = np.array(frame_timepoints)
    frame_signals    = np.array(frame_signals)
    frame_timepoints = frame_timepoints - 3 # A slight shift of the timepoints
                                            # to include the begginning of the peaks.

    error_check(frame_timepoints)

    return frame_timepoints, frame_signals

def reverse_detection(data, frame_timepoints, low_threshold, increment, precision=.95):
    """Detect frames in the left direction."""
    new_timepoints = []
    new_signals = []

    safe_increment = int(increment * (1+(1-precision)))

    i = frame_timepoints[0]-safe_increment
    while i>0:
        data_slice = data[i:i+increment//2+(safe_increment-increment)*2]
        if np.any(data_slice > low_threshold):
            i = i+np.argmax(data_slice > low_threshold)
        else:
            break #No low threshold crossing found -> no more frames to detect
        new_timepoints.append(i)
        i-= safe_increment #We move backward of almost a frame

    return new_timepoints[::-1]

def extend_timepoints(frame_timepoints, n=10):
    """Extrapolates points to the left. Not really needed now except for the signals idx that would change
    otherwise (and some starting index were set manually)"""
    frame_timepoints = np.array(frame_timepoints)
    typical_distance = int(np.mean(np.diff(frame_timepoints)))
    extended_tp = [frame_timepoints[0]-(i+1)*typical_distance for i in range(n) if (frame_timepoints[0]-(i+1)*typical_distance)>0]
    return extended_tp[::-1]

def error_check(frame_tp):
    """Search error by looking at the time between each frame.
    DLP is regular and odd time reveal misdetections."""
    deriv_frame_tp = np.diff(frame_tp)
    error_len_th = np.mean(deriv_frame_tp)+np.std(deriv_frame_tp)*6

    error_frames = np.abs(deriv_frame_tp)>error_len_th
    if np.any(error_frames):
        print("Error in timepoints detected in frames", np.where(error_frames)[0],
              "at timepoint", frame_tp[np.where(error_frames)[0]])

# Cell
def cluster_frame_signals(data, frame_timepoints, n_cluster=5):
    """Cluster the `frame_timepoints` in `n_cluster` categories depending on the area under the curve.
        - data: raw data used to compute the AUC
        - frame_timepoints: timepoints delimitating each frame
        - n_cluster: Number of cluster for the frame signals"""
    frame_aucs = np.fromiter(map(np.trapz, np.split(data, frame_timepoints)), float)
    if frame_timepoints[0] != 0: #We need to remove the first part if it wasn't a full frame
        frame_aucs = frame_aucs[1:]
    frame_auc_sorted = np.sort(frame_aucs)
    deriv = np.array(frame_auc_sorted[1:]-frame_auc_sorted[:-1])
    deriv[:5]  = 0 #removing tails values that can show weird stuff
    deriv[-5:] = 0
    threshold_peak = np.std(deriv)*3
    n          = n_cluster - 1
    idx_gaps = np.zeros(n+3, dtype="int")
    tmp_deriv = deriv.copy()
    zero_set_range = 10#int(len(deriv)*0.05) #Around the peaks, we set the values to 0 around
    for i in range(n+3): #Detecting more peaks than needed and then taking them starting on the right
        if tmp_deriv[np.argmax(tmp_deriv)] < threshold_peak:
            if i<n_cluster-1:
                print("Less transition in AUC detected than needed, results will be weird")
            break
        idx_gaps[i] = np.argmax(tmp_deriv)
        tmp_deriv[idx_gaps[i]-zero_set_range:idx_gaps[i]+zero_set_range] = 0
    idx_gaps = np.sort(idx_gaps)
    idx_gaps = idx_gaps[-(n_cluster-1):]
    thresholds = np.zeros(n, dtype="float")
    for i, idx in enumerate(idx_gaps):
        thresholds[i] = (frame_auc_sorted[idx+1] + frame_auc_sorted[idx])/2

    return np.array([np.sum(auc>thresholds) for auc in frame_aucs], dtype=int)

def cluster_by_epochs(data, frame_timepoints, frame_signals, epochs):
    """Does the same thing as `cluster_frame_signals`, but working on epochs around which the
    number of cluster can differ. Useful when a record contains stimuli with different signals sizes."""

    frame_aucs = np.fromiter(map(np.trapz, np.split(data, frame_timepoints)), float)
    if frame_timepoints[0] != 0: #We need to remove the first part if it wasn't a full frame
        frame_aucs = frame_aucs[1:]

    max_cluster = max([nclust-1 for (_,_,nclust) in epochs])

    for start,stop,n_cluster in epochs:
        n          = n_cluster - 1
        norm_clust = max_cluster/n
        frame_auc_sorted = np.sort(frame_aucs[start:stop])
        deriv = np.array(frame_auc_sorted[1:]-frame_auc_sorted[:-1])
        deriv[:5]  = 0 #removing tails values that can show weird stuff
        deriv[-5:] = 0
        idx_gaps   = np.zeros(n, dtype="int")
        tmp_deriv = deriv.copy()
        zero_set_range = 10
        for i in range(n):
            idx_gaps[i] = np.argmax(tmp_deriv)
            tmp_deriv[idx_gaps[i]-zero_set_range:idx_gaps[i]+zero_set_range] = 0
        idx_gaps = np.sort(idx_gaps)
        thresholds = np.zeros(n, dtype="float")
        for i, idx in enumerate(idx_gaps):
            thresholds[i] = (frame_auc_sorted[idx+1] + frame_auc_sorted[idx])/2

        frame_signals[start:stop] = np.array([np.sum(auc>thresholds)*norm_clust for auc in frame_aucs[start:stop]], dtype=int)
    return frame_signals

# Cell
def cluster_by_list(data, frame_timepoints, frame_signals, stim_list):
    """Assign the stimulus identity values from stim_list to the frames in data. stim_list contains only the
        sequence of stimuli. Those need to be expanded. Opposed to cluster_frame_signals and cluster_by_epochs no
        AUC operation is performed.
        Input:
            - data: raw data used to compute the stimulus times
            - frame_timepoints: timepoints delimiting each frame
            - frame_signals: binary 1-D numpy array indicating if high_threshold was passed in 'detect_frames'
            - stim_list: 1-D numpy array containing the sequence of the stimuli presented
        Output:
            - frame_signals: [same size as frame_timepoints] stim_signals list containing the correct value from
                stim_list at every entry"""
    # Determine the stimulus on- and offsets and their location relative to data
    stim_change = np.where(frame_signals[:-1] != frame_signals[1:])[0]
    stim_change = stim_change + 1 # since I need to compare to [1:] all values are shifted by 1
    #stim_idx = frame_timepoints[stim_change]

    # QDSpy currently is set to emit a short peak at the end to indicate the end of the stimulus presentation
    # This peak needs to be ignored
    epoch_end = stim_change[-2:] # return it for future analysis
    stim_change = stim_change[:-2] #add it to the no stimulus category

    # Split into on times & values vs off times & values
    stim_ons = stim_change[0::2]
    #stim_ons_idx = stim_idx[0::2]
    stim_offs = stim_change[1::2]
    #stim_offs_idx = stim_idx[1::2]

    # Replace the frame_signal entries with the stimulus codes
    frame_signals[frame_signals == 0] = -1 # To avoid confusion with the '0' stimulus code
    for i,stim_type in enumerate(stim_list):
        frame_signals[stim_ons[i]:stim_offs[i]] = stim_type

    return frame_signals, stim_ons, stim_offs, epoch_end

# Cell
def parse_time(time_str, pattern="%y%m%d_%H%M%S"):
    """Default parser of rhd timestamps. (serve as a template too)"""
    return datetime.datetime.strptime(time_str, pattern)

def get_position_estimate(stim_time, record_time, sampling_rate):
    """Estimate where in the record should a stimulus start, in sample points"""
    if stim_time < record_time:
        return -1
    else:
        return (stim_time - record_time).seconds * sampling_rate

# Cell
def match_starting_position(frame_timepoints, frame_signals, stim_signals, estimate_start, search_size=1000):
    """
    Search the best matching index between frame_signals and stim_signals.
    params:
        - frame_timepoints: Indexes of the frames in the record
        - frame_signals: Signals of the detected frames
        - stim_signals: Expected stimulus signals
        - estimate_start: Estimated start index
        - search_size: Stimulus is searched in frame_signals[idx_estimate-search_size: idx_estimate+search_size]
    return:
        - best match for the starting position of the stimulus
    """
    stim_matching_len = min(600, np.where(np.diff(stim_signals)!=0)[0][50]) #Way of getting the 50th change in the signals
    idx_estimate = np.argmax(frame_timepoints>estimate_start)
    search_slice = slice(max(0, idx_estimate-search_size-frame_signals.idx), min(idx_estimate+search_size-frame_signals.idx, len(frame_signals)))
    return search_slice.start + np.argmax(np.correlate(frame_signals[search_slice],
                                                       stim_signals[:stim_matching_len]))


# Cell
def display_match(match_position, reference=None, recorded=None, corrected=None, len_line=50):
    start, mid, end = 0, len(reference)//2, len(reference)-len_line
    for line in [start, mid, end]:
        if reference is not None:
            print("REF ["+str(line)+"] "," ".join(map(str,map(int, reference[line:line+len_line]))))
        if recorded is not None:
            print("REC ["+str(line)+"] "," ".join(map(str,map(int, recorded[line+match_position:line+len_line+match_position]))))
        if corrected is not None:
            print("COR ["+str(line)+"] "," ".join(map(str,map(int, corrected[line:line+len_line]))))
        print()

# Cell
def frame_error_correction(signals, unpacked, algo="nw", **kwargs):
    """Correcting the display stimulus frame values. Shifts are first detected with one of
    `shift_detection_conv` or `shift_detection_NW` and applied to the stimulus template. Then single frame
    mismatch are detected and corrected.
        - signals: true signal values recorded
        - unpacked: stimulus tuple (inten,marker,shader)
        - algo: algorithm for shift detection among [nw, conv]
        - **kwargs: extra parameter for shift detection functions

        returns: stim_tuple_corrected, shift_log, (error_frames_idx, replacement_idx)"""

    if algo=="no_shift":
        intensity, marker, shader  = unpacked[0].copy(), unpacked[1].copy(), unpacked[2]
        if shader is not None:
            shader = shader.copy()
        error_frames, replacements = error_frame_matches(signals, marker, range_=5)
        shift_log = []
    else:
        if algo=="nw":
            shift_log = shift_detection_NW(signals.astype(int), unpacked[1].astype(int), **kwargs)
        elif algo=="conv":
            shift_log = shift_detection_conv(signals.astype(int), unpacked[1].astype(int), range_=5, **kwargs)
        intensity, marker, shader = apply_shifts(unpacked, shift_log)
        error_frames, replacements = error_frame_matches(signals, marker, range_=5)
    if len(error_frames)>0:
        intensity[error_frames]    = intensity[replacements]
        marker[error_frames]       = marker[replacements]
        if shader is not None:
            shader[error_frames] = shader[replacements]
    return (intensity, marker, shader), shift_log, list(zip(map(int,error_frames), map(int,replacements)))

def error_frame_matches(signals, marker, range_):
    """Find the frames mismatching and finds in the record the closest frame with an identical signal value"""
    error_frames = np.nonzero(signals!=marker)[0]
    if len(error_frames)>0:
        where_equal = [((np.where(marker[err_id-range_:err_id+(range_+1)] == signals[err_id])[0]) - range_) for err_id in error_frames]
    #filtering out the frames where no match was found
        tmp    = np.array([[wheq,err] for (wheq, err) in zip(where_equal, error_frames) if len(wheq)>0], dtype="object")
        if len(tmp)==0:
            replacements = np.empty(shape=(0,), dtype=int)
            error_frames = np.empty(shape=(0,), dtype=int)
        else:
            where_equal  = tmp[:,0]
            error_frames = tmp[:,1]

        #Choosing among the equal frame signals the one that is the closest
            closest_equal = [wheq[(np.abs(wheq)).argmin()] for wheq in where_equal]
            error_frames  = np.array(error_frames, dtype=int)
            replacements  = error_frames + np.array(closest_equal, dtype=int)
    else:
        replacements = np.empty(shape=(0,), dtype=int)
        error_frames = np.empty(shape=(0,), dtype=int)

    return error_frames, replacements

def apply_shifts(unpacked, op_log):
    """Applies the shifts found by either shift_detection functions"""
    inten, marker, shader = unpacked[0].copy(), unpacked[1].copy(), unpacked[2]
    if shader is not None:
        shader = shader.copy()
    orig_len = len(marker)
    for idx, op in op_log:
        if op=="ins": #We insert a frame
            marker = np.insert(marker, idx, marker[idx], axis=0)
            inten  = np.insert(inten , idx, inten[idx], axis=0)
            if shader is not None:
                shader = np.insert(shader, idx, shader[idx], axis=0)
        elif op=="del": #We concatenate without the deleted frame
            marker = np.concatenate((marker[:idx],marker[idx+1:]))
            inten  = np.concatenate((inten[:idx],inten[idx+1:]))
            if shader is not None:
                shader = np.concatenate((shader[:idx],shader[idx+1:]))
    marker = marker[:orig_len]
    inten  = inten[:orig_len]
    if shader is not None:
        shader = shader[:orig_len]
    return (inten, marker, shader)

def shift_detection_conv(signals, marker, range_):
    """Detect shifts with a convolution method. First look at how far the next closest frame are, and average
    it over the record. When the average cross the -1 or 1 threshold, shift the reference accordingly."""
    marker = marker.copy()
    shift_detected = True
    shift_log = []
    while shift_detected:
        error_frames, replacements = error_frame_matches(signals, marker, range_)

        all_shifts = np.zeros(len(marker))
        all_shifts[error_frames] = replacements-error_frames
        all_shifts_conv = np.convolve(all_shifts, [1/20]*20, mode="same") #Averaging the shifts to find consistant shifts

        shift_detected = np.any(np.abs(all_shifts_conv)>.5)
        if shift_detected: #iF the -.5 threshold is crossed, we insert a "fake" frame in the reference and we repeat the operation
            change_idx = np.argmax(np.abs(all_shifts_conv)>.5)
            if all_shifts_conv[change_idx]>.5:#Need to delete frame in reference
                #Need to refine index to make sure we delete a useless frame
                start,stop = max(0,change_idx-2), min(len(marker),change_idx+2)
                for i in range(start,stop):
                    if marker[i] not in signals[start:stop]:
                        change_idx = i
                        break
                shift_log.append([int(change_idx), "del"])
                marker = np.concatenate((marker[:change_idx], marker[change_idx+1:], [0]))
            else:#Need to insert frame in reference
                shift_log.append([int(change_idx), "ins"])
                #inserting a frame and excluding the last frame to keep the references the same length
                marker     = np.insert(marker, change_idx, marker[change_idx], axis=0)[:-1]
    return shift_log

def shift_detection_NW(signals, marker, simmat_basis=[1,-1,-3,-3,-1], insdel=-10, rowside=20):
    """Memory optimized Needleman-Wunsch algorithm.
    Instead of an N*N matrix, it uses a N*(side*2+1) matrix. Indexing goes slightly differently but
    result is the same, with far less memory consumption and exection speed scaling better with
    size of the sequences to align."""
    #Setting the similarity matrix
    side = rowside
    sim_mat = np.empty((len(marker), side*2+1), dtype="int32")
    #Setting the errors
    insertion_v = insdel #insertions are commons not so high penalty
    deletion_v  = insdel #deletions detection happens during periods of confusion but are temporary. High value
    error_match = np.array(simmat_basis) #The value for a 0 matching with [0,1,2,3,4]
    error_mat = np.empty((len(simmat_basis),len(simmat_basis)))
    for i in range(len(simmat_basis)):
        error_mat[i] = np.roll(error_match,i)

    #Filling the similarity matrix
    sim_mat[0, side] = error_mat[marker[0], signals[0]]
    #Initialization: Setting the score of the first few row and first few column cells
    for j in range(side+1, side*2+1):
        sim_mat[0,j] = sim_mat[0,side] + insertion_v*j
    for i in range(1, side+1):
        sim_mat[i,side-i] = sim_mat[0,side] + deletion_v*i

    #Corpus: if j is the first cell of the row, the insert score is set super low
    #        if j is the last  cell of the row, the delete score is set super low
    for i in range(1, sim_mat.shape[0]):
        start = max(side-i+1, 0)
        stop  = min(side*2+1, side+sim_mat.shape[0]-i)
        for j in range(start, stop):
            if j==0:#j==start and i>side:
                insert = -99999
                delete = sim_mat[i-1, j+1] + deletion_v
            elif j==side*2:
                delete = -99999
                insert = sim_mat[i, j-1] + insertion_v
            else:
                insert = sim_mat[i, j-1] + insertion_v
                delete = sim_mat[i-1, j+1] + deletion_v
            match  = sim_mat[i-1, j] + error_mat[marker[i], signals[j+i-side]]
            sim_mat[i,j] = max(insert,delete,match)

    #Reading the similarity matrix
    #In general, it's the same, at the difference that when i decrement, must add 1 to j compared to usual.
    i = len(marker)-1
    j = side
    shift_log = []
    while (i > 0 or j>side-i):
        if (i > 0 and j>side-i and sim_mat[i,j]==(sim_mat[i-1,j]+error_mat[marker[i], signals[j+i-side]])):
            i -= 1
        elif(i > 0 and sim_mat[i,j] == sim_mat[i-1,j+1] + deletion_v):
            shift_log.insert(0,(j+i-side+1, "del")) #Insert the j value for deletion too because all shifts
            i-=1                                        #are relative to the signals recorded, unlike normal NW
            j+=1
        else:
            shift_log.insert(0,(j+i-side, "ins"))
            j-=1

    return shift_log

# Cell
def chop_stim_edges(first_frame, last_frame, stim_tuple, shift_log, frame_replacement):
    """Cut out the stimulus parts not containing actual stimulus, and change the idx values of `shift_log`
    and `frame_replacement` to match the new indexing."""
    inten, marker, shader = stim_tuple
    if last_frame<0: #Using negative indexing
        last_frame = len(marker)+last_frame
    inten = inten[first_frame:last_frame]
    marker = marker[first_frame:last_frame]
    if shader is not None:
        shader = shader[first_frame:last_frame]

    shift_log = [(shift[0]-first_frame, shift[1]) for shift in shift_log if shift[0]<last_frame]
    frame_replacement = [(fr[0]-first_frame, fr[1]-first_frame) for fr in frame_replacement if fr[0]<last_frame]

    return (inten, marker, shader), shift_log, frame_replacement

# Cell
def detect_calcium_frames(scanning_data, epoch_threshold=-8):
    """Detect the timing of the 2P frames, epoch by epoch over a record."""

    #Finds the start of a stack recording
    start_set = np.where((scanning_data[1:] > epoch_threshold) & (scanning_data[:-1] < epoch_threshold))[0]
    #Finds the end of a stack recording
    end_set   = np.where((scanning_data[1:] < epoch_threshold) & (scanning_data[:-1] > epoch_threshold))[0]
    #Splits the records into the epochs
    list_epoch = np.array_split(scanning_data, np.ravel(list(zip(start_set, end_set))))[1::2]

    def detect_peak_sync(epoch):
        #Finds the peaks in an epoch. Peaks have strong SNR so this works fine
        return signal.find_peaks(epoch, prominence=2)[0]

    return [arr + start_set[i] for i, arr in enumerate(list(map(detect_peak_sync, list_epoch)))]


# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/theonerig/',
                'doc_host': 'https://Tom-TBT.github.io',
                'git_url': 'https://github.com/Tom-TBT/theonerig/tree/master/',
                'lib_path': 'theonerig'},
  'syms': { 'theonerig.core': {},
            'theonerig.database': {},
            'theonerig.eyetrack': {},
            'theonerig.leddome': {},
            'theonerig.modelling': {},
            'theonerig.plotting': {},
            'theonerig.processing': {},
            'theonerig.synchro.extracting': {},
            'theonerig.synchro.io': {},
            'theonerig.synchro.nested_stims': {},
            'theonerig.synchro.processing': {},
            'theonerig.testdata': {},
            'theonerig.utils': {}}}


# AUTOGENERATED! DO NOT EDIT! File to edit: 05_database.ipynb (unless otherwise specified).

__all__ = ['get_db_engine', 'prompt_credentials', 'get_record_essentials', 'get_stim_params', 'get_table',
           'stim_param_to_dict']

# Cell
from sqlalchemy import create_engine, text
from sqlalchemy import Table, Column, Integer, String, MetaData, select

import pandas as pd
import getpass
import json

# Cell
def get_db_engine(username, password, ip_adress, model_name, rdbms="mysql"):
    """
    Creates a sqlalchemy engine to query a database.

    params:
        - username: Username used to connect
        - password: Password of the user
        - ip_adress: IP adress of the database
        - model_name: Name of the model of the database to connect to
        - rdbms: Backend engine of the database

    return:
        - A sqlalchemy engine connected to the database
    """
    engine = create_engine("%s://%s:%s@%s/%s" % (rdbms, username, password, ip_adress, model_name),echo = False)
    test_query = "SELECT * FROM Project"
    df_test = pd.DataFrame(engine.connect().execute(text(test_query)))
    return engine

def prompt_credentials(user=None, db_adress=None):
    """
    Helper function to make a prompt for the password, and additonally the user and the database IP adress
    if left to None.

    params:
        - user: None to prompt or name of the user.
        - db_adress: None to prompt or database adress

    return:
        - username, password and database IP adress
    """
    if user is None:
        user = input(prompt='Username: ')
    passwd = getpass.getpass(prompt='Password: ')
    if db_adress is None:
        db_adress = input(prompt='DB IP: ')
    return user, passwd, db_adress

# Cell
def get_record_essentials(engine, record_id):
    """
    Retrieves the essential informations about a record.

    params:
        - engine: Database engine
        - record_id: ID of the record

    return:
        - Pandas Dataframe of record essential informations
    """
    q_select_record = "SELECT * FROM Record WHERE id = %d" % record_id
    q_select_cell = "SELECT * FROM Cell WHERE record_id = %d" % record_id

    df_record = pd.DataFrame(engine.connect().execute(text(q_select_record)))
    df_cell = pd.DataFrame(engine.connect().execute(text(q_select_cell)))

    experiment_id = df_record["experiment_id"][0]
    q_select_experiment = "SELECT * FROM Experiment WHERE id = %d" % experiment_id
    df_experiment = pd.DataFrame(engine.connect().execute(text(q_select_experiment)))

    mouse_id = df_experiment["mouse_id"][0]
    q_select_mouse = "SELECT * FROM Mouse WHERE id = %d" % mouse_id
    df_mouse = pd.DataFrame(engine.connect().execute(text(q_select_mouse)))

    tool_id = df_record["tool_id"][0]
    q_select_tool = "SELECT * FROM Tool WHERE id = %d" % tool_id
    df_tool = pd.DataFrame(engine.connect().execute(text(q_select_tool)))

    q_select_map = "SELECT * FROM Map WHERE tool_id = %d" % tool_id
    df_map = pd.DataFrame(engine.connect().execute(text(q_select_map)))

    res_dict = {"record": df_record, "cell": df_cell,
                "experiment": df_experiment, "mouse": df_mouse,
                "tool": df_tool, "map": df_map}
    return res_dict

# Cell
def get_stim_params(engine, stim_hashes):
    """
    Retrieves the parameters of a stimulus specified by its hash key.

    params:
        - engine: Database engine
        - stim_hashes: Stimulus hash

    return:
        - Pandas Dataframe of stimulus parameters
    """
    #Writting the query speed up the function rather than querying all individual tables
    # and filtering them all
    if not isinstance(stim_hashes, list):
        stim_hashes = [stim_hashes]
    if len(stim_hashes)==1:
        str_hashes = "('"+stim_hashes[0]+"')"
    else:
        str_hashes=str(tuple(stim_hashes))
    query = """SELECT Stim.name AS stim_name, description, barcode, stim_comment, stimulus_id,
                screen_id, hash, date AS date_compiled, compiled_comment, compiled_id, parameter_id,
                Parameter.name as param_name, value as param_value
                FROM (SELECT Compiled.id as comp_id, name, description, barcode, Stimulus.comment AS stim_comment, stimulus_id, screen_id, hash, date, Compiled.comment AS compiled_comment FROM Stimulus
                LEFT JOIN Compiled ON stimulus_id=Stimulus.id WHERE hash IN """+str_hashes+""") AS Stim
                LEFT JOIN Compiled_Parameter ON compiled_id = comp_id
                LEFT JOIN Parameter ON parameter_id = Parameter.id"""
    df_params = pd.DataFrame(engine.connect().execute(text(query)))
    return df_params

# Cell
def get_table(engine, table_name):
    """
    Return the entire content of a table in a pandas Dataframe.

    params:
        - engine: Database engine
        - table_name: Name of the table

    return:
        - Pandas Dataframe of the whole table
    """
    query = """SELECT * FROM """+str(table_name)
    df_table = pd.DataFrame(engine.connect().execute(text(query)))
    return df_table

# Cell
def stim_param_to_dict(param_df, md5):
    param_dict = {}
    stim_mask = param_df["hash"] == md5
    for _, row in param_df[stim_mask][["param_name", "param_value"]].iterrows():
        try:
            param = json.loads(row["param_value"])
        except:
            param = row["param_value"]
        param_dict[row["param_name"]] = param
    return param_dict


# AUTOGENERATED! DO NOT EDIT! File to edit: 13_leddome.ipynb (unless otherwise specified).

__all__ = ['get_dome_positions', 'as_cartesian', 'as_spherical', 'angular_distance', 'build_wave_stimulus_array',
           'Quaternion', 'get_waves_relative_position', 'get_led_relative_position']

# Cell
import numpy as np
import math
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# Cell
def get_dome_positions(mode="cartesian"):
    """
    Generates positions of all LEDs of the dome. Position of first and last LED of each stripe
    were estimated in Blender, and other in between LEDs position are interpolated from those two.

    params:
        - mode: What coordinates to obtain in set ["cartesian", "spherical"]
    return:
        - LED position of the LED dome, in shape (4, 237). shape[0] organized by quarter (q1 to q4) and shape[1] is the
        concatenation of (left,right)
    """
    assert mode in ["cartesian", "spherical"], """Mode must be one of ["cartesian", "spherical"]"""
    stripe_dict = {}
    stripe = np.array([[-0.44162,0.46045,10.07932], [-0.03378,10.07122,0.72211]])*10
    stripe_dict["A"] = _slerp(stripe, 23)

    stripe = np.array([[0.42254,1.33094,10.00507], [0.83062,9.99418,1.12168]])*10
    stripe_dict["B"] = _slerp(stripe, 21)

    stripe = np.array([[-1.3044,1.33575,9.94323], [-0.93444, 10.00996,0.99274]])*10
    stripe_dict["C"] = _slerp(stripe, 21)

    stripe = np.array([[1.35075,2.2321,9.75535], [1.68846,9.91944,0.77928]])*10
    stripe_dict["D"] = _slerp(stripe, 20)

    stripe = np.array([[-2.20708,2.29345,9.58381], [-1.8337,9.92046,1.14081]])*10
    stripe_dict["E"] = _slerp(stripe, 19)

    stripe = np.array([[2.31814,3.13993,9.31365], [2.52401,9.74959,0.86306]])*10
    stripe_dict["F"] = _slerp(stripe, 18)

    stripe = np.array([[-3.15667,3.31007,9.00523], [-2.69219,9.68376,1.0918]])*10
    stripe_dict["G"] = _slerp(stripe, 17)

    stripe = np.array([[3.3186,4.12493,8.60008], [3.28828,9.52856,0.61278]])*10
    stripe_dict["H"] = _slerp(stripe, 16)

    stripe = np.array([[-4.0779,4.27888,8.18478], [-3.45295,9.45243,0.77226]])*10
    stripe_dict["I"] = _slerp(stripe, 15)

    stripe = np.array([[4.29328,5.00709,7.63564], [4.17924,9.14635,1.03659]])*10
    stripe_dict["J"] = _slerp(stripe, 13)

    stripe = np.array([[-4.99026,5.24451,7.06361], [-4.3501,9.07599,1.00064]])*10
    stripe_dict["K"] = _slerp(stripe, 12)

    stripe = np.array([[5.22638,5.86208,6.3335], [4.85207,8.84847,0.57339]])*10
    stripe_dict["L"] = _slerp(stripe, 11)

    stripe = np.array([[-5.77797,6.10141,5.60405], [-5.14097,8.63676,1.02421]])*10
    stripe_dict["M"] = _slerp(stripe, 9)

    stripe = np.array([[6.03059,6.57628,4.71668], [5.55174,8.42348,0.46679]])*10
    stripe_dict["N"] = _slerp(stripe, 8)

    stripe = np.array([[-6.40277,6.82204,3.80993], [-5.84937,8.19519,0.84915]])*10
    stripe_dict["O"] = _slerp(stripe, 6)

    stripe = np.array([[6.62294,7.08816,2.77088], [6.34649,7.81552,0.85683]])*10
    stripe_dict["P"] = _slerp(stripe, 4)

    stripe = np.array([[-6.77734,7.27747,1.7878], [-6.49463,7.71771,0.6162]])*10
    stripe_dict["Q"] = _slerp(stripe, 3)

    stripe = np.array([[6.94329,7.30411,0.65871]])*10
    stripe_dict["R"] = stripe

    res = _symetry_stripes(_chain_stripes(stripe_dict))
    if mode=="spherical":
        res = np.apply_along_axis(as_spherical, axis=-1, arr=res)
    return res

def _symetry_stripes(stripe):
    """
    Generates the 90° symetry of three stripes from the given stripe.
    """
    all_stripes = np.stack([stripe]*4, axis=0)
    tmp = all_stripes[1,:,0]*-1
    all_stripes[1,:,0] = all_stripes[1,:,1]
    all_stripes[1,:,1] = tmp

    all_stripes[2,:,0] *= -1
    all_stripes[2,:,1] *= -1

    tmp = all_stripes[3,:,1]*-1
    all_stripes[3,:,1] = all_stripes[3,:,0]
    all_stripes[3,:,0] = tmp
    return all_stripes

def _slerp(leds_xyz, n_led):
    """Interpolate positions from the xyz positon of the first and last LED

    params:
        -leds_xyz: np.array of shape(2,3)
        -n_led: total n LED on the stripe
    return:
        - interpolated positions
    """
    p0, p1 = leds_xyz[0], leds_xyz[1]

    omega = math.acos(np.dot(p0/np.linalg.norm(p0), p1/np.linalg.norm(p1)))
    so = math.sin(omega)
    return [math.sin((1.0-t)*omega) / so * p0 + math.sin(t*omega)/so * p1 for t in np.linspace(0.0, 1.0, n_led)]

def as_cartesian(rthetaphi, is_radian=True):
    """
    Convert 3D polar coordinate tuple into cartesian coordinates.

    params:
        - rthetaphi: Single or list of (r, theta, phi) iterable
        - is_radian: Boolean to specify if in radians or in degrees
    return:
        - Single or list of converted (x, y, z) array.
    """
    r, theta, phi = tuple(np.array(rthetaphi).T)
    if not is_radian:
        theta   = theta*np.pi/180
        phi     = phi*np.pi/180
    x = r * np.sin( theta ) * np.cos( phi )
    y = r * np.sin( theta ) * np.sin( phi )
    z = r * np.cos( theta )
    return np.stack([x,y,z], axis=-1)

def as_spherical(xyz):
    """
    Convert 3D cartesian coordinates tuple into polar coordinate.

    params:
        - xyz: Single or list of (x, y, z) iterable
    return:
        - Single or list of converted (r, theta, phi) array.
    """
    x, y, z = tuple(np.array(xyz).T)
    r       =  np.sqrt(x*x + y*y + z*z)
    theta   =  np.arccos(z/r)
    phi     =  np.arctan2(y,x)
    return np.stack([r,theta,phi], axis=-1)

def angular_distance(theta_1, phi_1, theta_2, phi_2):
    """
    Computes the angle separating two points (or a list of points) on a sphere, in radians.
    params:
        - theta_1: Theta angle of the first point.
        - phi_1:   Phi angle of the first point.
        - theta_2: Theta angle of the second point.
        - phi_2:   Phi angle of the second point.
    """
    theta_1 = (np.pi/2)-theta_1 #The formula works for Declination angle
    theta_2 = (np.pi/2)-theta_2

    return np.arccos(np.sin(theta_1)*np.sin(theta_2) +
                     np.cos(theta_1)*np.cos(theta_2)*np.cos(phi_2-phi_1))

def _chain_stripes(stripe_dict):
    """
    Chain the stripes to create a one-dimensional array were LED idx correspond to their index on the stripe,
    with left side first.
    """
    res = []
    UP,DOWN = -1,1
    ori = UP
    left_side = ["B","D","F","H","J","L","N","P","R"]
    for key in left_side:
        res.extend(stripe_dict[key][::ori])
        ori *= -1

    ori = UP
    right_side = ["Q","O","M","K","I","G","E","C","A"]
    for key in right_side:
        res.extend(stripe_dict[key][::ori])
        ori *= -1
    return np.array(res)

# Cell
def build_wave_stimulus_array(epoch_sequence, wave_width=0.58, wave_speed=.58, n_frame_epoch=640, n_frame_isi=50, frame_rate=100):
    """
    Build the numpy stimulus matrix of the LED values for each frame.
    params:
        - epoch_sequence: Sequence of indexes played randomly during the stimulation
        - wave_width: Width of the wave in radians
        - wave_speed: Speed of the wave in radians.s-1
        - n_frame_epoch: Number of frames that an epoch last
        - n_frame_isi: Number of frames during teh inter-stimulus-interval
        - frame_rate: Frame rate of the display
    returns:
        - LED values for the wave stimulus, in shape (t, 4, 237), where t=n_epoch*(n_frame_epoch+n_frame_isi)
    """
    n_epoch = np.max(epoch_sequence)+1

    polar_pos  = get_dome_positions(mode="spherical") #r, theta, phi
    theta_leds = polar_pos[:,:,1].reshape(-1)
    phi_leds   = polar_pos[:,:,2].reshape(-1)

    indexes  = np.arange(n_epoch)+0.5
    #Theta and phi make the axis around which the LEDs are rotated by alpha
    theta    = np.pi/2; #Theta is fixed, corresponds to the plane touching the dome edge (elevation=0° or inclination=90°)
    phis     = np.pi*(1 + np.sqrt(5)) * indexes #Angle of rotation around the centre
    alphas   = np.arccos(1 - 2*indexes/n_epoch)     #Distance angle from the centre

    tmp  = np.sin(0.5*alphas)
    qA_0 = np.cos(0.5*alphas)[:, None] #Quaternion of the rotation. Adding axis to do matrix multiplication
    qA_x = (np.sin(theta)*np.cos(phis)*tmp)[:, None]
    qA_y = (np.sin(theta)*np.sin(phis)*tmp)[:, None]
    qA_z = (np.cos(theta)         *tmp)[:, None]

    qB_x_leds = (np.sin(theta_leds)*np.cos(phi_leds))[:, None].T
    qB_y_leds = (np.sin(theta_leds)*np.sin(phi_leds))[:, None].T
    qB_z_leds = (np.cos(theta_leds))[:, None].T

    #Computes each LED elevation for each rotation. Involve Quaternions. Only the relevant values are calculated
    mq0 = -(qA_x@qB_x_leds) - (qA_y@qB_y_leds) - (qA_z@qB_z_leds)
    mqx =  (qA_0@qB_x_leds) + (qA_y@qB_z_leds) - (qA_z@qB_y_leds)
    mqy =  (qA_0@qB_y_leds) - (qA_x@qB_z_leds) + (qA_z@qB_x_leds)
    mqz =  (qA_0@qB_z_leds) + (qA_x@qB_y_leds) - (qA_y@qB_x_leds)

    mqz = (mq0 * (-qA_z)) + (mqx * (-qA_y)) - (mqy * (-qA_x)) + (mqz * qA_0)

    LED_elevations = np.arccos(mqz)
    LED_elevations = LED_elevations[epoch_sequence]

    frame_step      = wave_speed/frame_rate
    wave_elevations = np.arange(n_frame_epoch)*frame_step #Building the values used to compare the LED_elevation with

    time_shape = n_epoch*(n_frame_epoch+n_frame_isi)
    result     = np.empty((time_shape, *theta_leds.shape))
    for i, LED_elevation in enumerate(LED_elevations):
        for j, wave_elevation in enumerate(wave_elevations):
            result[i*(n_frame_epoch+n_frame_isi)+j] = (wave_elevation-wave_width<LED_elevation) & (LED_elevation<wave_elevation)
    return result.reshape(time_shape,4, -1)

# Cell
class Quaternion( object ):
    """
    Simplified Quaternion class for rotation of normalized vectors only!
    """

    def __init__( self, q0, qx, qy, qz ):
        """
        Internally uses floats to avoid integer division issues.

        @param q0: int or float
        @param qx: int or float
        @param qy: int or float
        @param qz: int or float
        """
        self._q0 = float( q0 )
        self._qx = float( qx )
        self._qy = float( qy )
        self._qz = float( qz )
        """
        Note if interpreted as rotation q0 -> -q0 doesn't make a difference
        q0 = cos( w ) so -cos( w ) = cos( w + pi ) and as the rotation
        is by twice the angle it is either 2w or 2w + 2pi, the latter being equivalent to the former.
        """

    def conjugate(q):
        """
        @return Quaternion
        """
        conjq = Quaternion( q._q0, -q._qx, -q._qy, -q._qz )
        return conjq

    def __mul__(q, r):
        """
        Non commutative quaternion multiplication.
        @return Quaternion
        """
        if isinstance(r, Quaternion):
            mq0 = q._q0 * r._q0 - q._qx * r._qx - q._qy * r._qy - q._qz * r._qz
            mqx = q._q0 * r._qx + q._qx * r._q0 + q._qy * r._qz - q._qz * r._qy
            mqy = q._q0 * r._qy - q._qx * r._qz + q._qy * r._q0 + q._qz * r._qx
            mqz = q._q0 * r._qz + q._qx * r._qy - q._qy * r._qx + q._qz * r._q0
            out = Quaternion(mq0, mqx, mqy, mqz)
        else:
            raise TypeError
        return out

    def __getitem__( q, idx ):
        """
        @return float
        """
        if idx < 0:
            idx = 4 + idx
        if idx in [ 0, 1, 2, 3 ]:
            out = (q._q0, q._qx, q._qy, q._qz)[idx]
        else:
            raise IndexError
        return out

# Cell
def get_waves_relative_position(cell_sta_position, n_waves=100, mode="spherical"):
    """
    Rotate the waves origins to obtain for the ref_led_flat_idx a spherical position of (0,0)
    params:
        - cell_sta_position: (theta, phi) Tuple of the cell position in spherical coordinates
        - n_waves: Number of waves in the wave stimulus (positions/density of waves determined by this parameter)
        - mode: One of ["spherical", "cartesian"], for the returned position
    return:
        - The rotated waves position
    """
    assert mode in ["spherical", "cartesian"], 'Mode must be one of ["spherical", "cartesian"]'

    theta_led = cell_sta_position[0]
    phi_led   = cell_sta_position[1]

    #Creation of the rotation quaternion and it's conjugate
    theta_rot    = np.pi/2; #Theta is fixed, corresponds to the plane touching the dome edge (elevation=0°)
    phi_rot      = phi_led+np.pi/2
    alpha_rot    = -theta_led

    xA, yA, zA = np.sin(theta_rot)*np.cos(phi_rot), np.sin(theta_rot)*np.sin(phi_rot), np.cos(theta_rot)
    tmp        = np.sin(0.5*alpha_rot)
    rot_quat   = Quaternion(np.cos(0.5*alpha_rot), xA*tmp, yA*tmp, zA*tmp)
    rot_conj   = rot_quat.conjugate()

    #Creation of the wave quaternions
    indexes       = np.arange(n_waves)+0.5
    phis_wave     = np.pi*(1 + np.sqrt(5)) * indexes + np.pi/2  #Angle of rotation around the centre. Add pi/2 to correspond to the displayed wave positon
    theta_wave    = np.arccos(1 - 2*indexes/n_waves)  #Distance angle from the centre
    xB, yB, zB = np.sin(theta_wave)*np.cos(phis_wave), np.sin(theta_wave)*np.sin(phis_wave), np.cos(theta_wave)
    quaternions_wave = [Quaternion(0,x,y,z) for x,y,z in zip(xB, yB, zB)]

    #Rotation of the waves
    rotated_waves      = [rot_quat*(q*rot_conj) for q in quaternions_wave]
    if mode=="spherical":
        return np.array([as_spherical((q[1], q[2], q[3])) for q in rotated_waves])
    else:
        return np.array([(q[1], q[2], q[3]) for q in rotated_waves])

def get_led_relative_position(ref_led_flat_idx, mode="spherical"):
    """
    Rotate the LED positions to obtain for the ref_led_flat_idx a spherical position of (0,0)
    params:
        - ref_led_flat_idx: Flattened index of the reference LED (e.g. obtained with np.argmax on abs STA values)
        - mode: One of ["spherical", "cartesian"], for the returned position
    return:
        - The rotated LED position
    """
    assert mode in ["spherical", "cartesian"], 'Mode must be one of ["spherical", "cartesian"]'

    theta_led = get_dome_positions(mode="spherical")[ref_led_flat_idx//237,ref_led_flat_idx%237,1]
    phi_led   = get_dome_positions(mode="spherical")[ref_led_flat_idx//237,ref_led_flat_idx%237,2]

    #Creation of the rotation quaternion and it's conjugate
    theta_rot    = np.pi/2; #Theta is fixed, corresponds to the plane touching the dome edge (elevation=0°)
    phi_rot      = phi_led+np.pi/2
    alpha_rot    = -theta_led

    xA, yA, zA = np.sin(theta_rot)*np.cos(phi_rot), np.sin(theta_rot)*np.sin(phi_rot), np.cos(theta_rot)
    tmp        = np.sin(0.5*alpha_rot)
    rot_quat   = Quaternion(np.cos(0.5*alpha_rot), xA*tmp, yA*tmp, zA*tmp)
    rot_conj   = rot_quat.conjugate()

    cart_pos  = get_dome_positions(mode="cartesian") #r, theta, phi
    x_leds = cart_pos[:,:,0].reshape(-1)
    y_leds = cart_pos[:,:,1].reshape(-1)
    z_leds = cart_pos[:,:,2].reshape(-1)

    quaternions_leds = [Quaternion(0,x,y,z) for x,y,z in zip(x_leds, y_leds, z_leds)]
    rotated_leds     = [rot_quat*(q*rot_conj) for q in quaternions_leds]
    if mode=="spherical":
        relative_led_pos = np.array([as_spherical((q[1], q[2], q[3])) for q in rotated_leds])
    else:
        relative_led_pos = np.array([(q[1], q[2], q[3]) for q in rotated_leds])

    return relative_led_pos


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_utils.ipynb (unless otherwise specified).

__all__ = ['extend_sync_timepoints', 'align_sync_timepoints', 'resample_to_timepoints', 'link_sync_timepoints',
           'flip_stimulus', 'flip_gratings', 'stim_to_dataChunk', 'phy_results_dict', 'spike_to_dataChunk',
           'get_calcium_stack_lenghts', 'twoP_dataChunks', 'img_2d_fit', 'fill_nan', 'stim_inten_norm',
           'group_direction_response', 'group_chirp_bumps', 'get_repeat_corrected', 'removeSlowDrift',
           'time_shift_test_corr', 'cross_corr_with_lag', 'get_inception_generator', 'group_omitted_epochs',
           'get_shank_channels', 'format_pval', 'stim_recap_df']

# Cell
import numpy as np
import pandas as pd
import os
import glob
import re
from typing import Dict, Tuple, Sequence, Union, Callable
import scipy.interpolate as interpolate
from scipy.ndimage import convolve1d
from scipy.signal import savgol_filter
import scipy.stats
from scipy.ndimage import gaussian_filter
import matplotlib.pyplot as plt
import math
from cmath import *
from PIL import Image

from .core import *

# Cell
def extend_sync_timepoints(timepoints:np.ndarray, signals:np.ndarray,
                           up_bound, low_bound=0) -> Tuple[DataChunk, DataChunk]:
    """
    Extend arrays of timepoints and signals (with identical shape) from the low_bound up to the up_bound.
    For example, the first timepoint could be 2000, and with a low_bound of 0, it would add the
    timepoints 0, 500, 1000, 1500 if the timepoint distance is of 500 (obtained by averaging the timepoints
    distances).

    params:
        - timepoints: Timepoints to extend
        - signals: Signals to extend
        - up_bound: Up bound to which to extend both timepoints and signals
        - low_bound: Low bound to which to extend both timepoints and signals

    returns:
        - timepoint: Extended timepoints
        - signals: The datachunk array is not modified, but the idx attribute is increased by the number
        of frames added with the low_bound.
    """
    assert len(timepoints) == len(signals)
    timepoints = np.array(timepoints)
    signals = np.array(signals)
    spb = np.mean(timepoints[1:]-timepoints[:-1]) #spf: sample_per_bin

    #Left and right side are just prolongation of the sample_times up
    # from (0-sample_per_fr) to (len+sample_per_fr) so it covers all timepoints
    left_side  = np.arange(timepoints[0]-spb , low_bound - spb, -spb)[::-1].astype(int)
    right_side = np.arange(timepoints[-1]+spb,  up_bound + spb,  spb).astype(int)

    new_timepoints = np.concatenate((left_side,
                                     timepoints,
                                     right_side))

    timepoint_chunk = DataChunk(data=new_timepoints, idx=0, group="sync")
    signal_chunk    = DataChunk(data=signals, idx=len(left_side), group="sync")
    return (timepoint_chunk, signal_chunk)

# Cell
def align_sync_timepoints(timepoints:DataChunk, signals:DataChunk,
                          ref_timepoints:DataChunk, ref_signals:DataChunk) -> Tuple[DataChunk, DataChunk, DataChunk]:
    """
    Align the signals of a timepoints timeserie to a reference ref_timepoints with the corresponding
    ref_signals. ref_timepoints is extended to match ref_timepoints lenght.

    params:
        - timepoints: timepoints to align
        - signals: signals to align
        - ref_timepoints: reference timepoints
        - ref_signals: reference signals

    return:
        - Aligned timepoints (DataChunk)
        - Aligned signals (DataChunk)
    """
    shift_left = ((np.where(ref_signals)[0][0] + ref_signals.idx)
                  - (np.where(signals)[0][0] + signals.idx))
    shift_right   = len(ref_timepoints) - (len(timepoints) + shift_left)

    spb     = np.mean(timepoints[1:]-timepoints[:-1]) #spf: sample_per_bin
    spb_ref = np.mean(ref_timepoints[1:]-ref_timepoints[:-1]) #spf: sample_per_bin

    left_timepoints      = np.zeros(0)
    left_timepoints_ref  = np.zeros(0)
    right_timepoints     = np.zeros(0)
    right_timepoints_ref = np.zeros(0)

    if shift_left > 0: #the ref started before, need to extend the other
        init  = timepoints[0]-spb
        left_timepoints = np.arange(init ,
                                    init-(spb*shift_left+1),
                                    -spb)[:shift_left][::-1].astype(int)
    else:
        shift_left = abs(shift_left)
        init  = ref_timepoints[0]-spb_ref
        left_timepoints_ref = np.arange(init ,
                                        init-(spb_ref*shift_left+1),
                                        -spb_ref)[:shift_left][::-1].astype(int)
        #We also need to shift the index of the ref signals since we increased the size of the ref_timepoints
        ref_signals.idx = ref_signals.idx + len(left_timepoints_ref)

    if shift_right > 0: #the ref ended after, need to extend the other
        init  = timepoints[-1]+spb
        right_timepoints = np.arange(init ,
                                    init+(spb*shift_right+1),
                                    spb)[:shift_right].astype(int)
    else:
        shift_right = abs(shift_right)
        init  = ref_timepoints[-1]+spb_ref
        right_timepoints_ref = np.arange(init ,
                                    init+(spb_ref*shift_right+1),
                                    spb_ref)[:shift_right].astype(int)

    timepoint    = DataChunk(data=np.concatenate((left_timepoints,
                                     timepoints,
                                     right_timepoints)), idx=0, group="sync")

    timepoint_ref = DataChunk(data=np.concatenate((left_timepoints_ref,
                                     ref_timepoints,
                                     right_timepoints_ref)), idx=0, group="sync")

    return (timepoint, timepoint_ref, ref_signals)

# Cell
def resample_to_timepoints(timepoints:np.ndarray, data:np.ndarray,
                             ref_timepoints:DataChunk, group="data") -> DataChunk:
    """
    Resample the data at timepoints to new timepoints given by ref_timepoints.
    Return a DataChunk of the resampled data belonging to a specified group.

    params:
        - timepoints: Original timepoints of the data
        - data: Data to resample of shape (t, ...)
        - ref_timepoints: Target timepoints for the resampling
        - group: Group assigned to the returned DataChunk

    return:
        - Resampled datachunk with appropriate idx.
    """

    assert len(timepoints) == len(data)
    timepoints = np.array(timepoints)
    data = np.array(data)

    start_idx = np.argmax(ref_timepoints >= timepoints[0])
    stop_idx  = np.argmax(ref_timepoints >= timepoints[-1])
    if stop_idx == 0:
        stop_idx = len(ref_timepoints)

    if len(ref_timepoints[start_idx:stop_idx]) < len(timepoints): #Downsampling
        distance = (np.argmax(timepoints>ref_timepoints[start_idx+1])
                - np.argmax(timepoints>ref_timepoints[start_idx]))

        kernel = np.ones(distance)/distance
        data = convolve1d(data, kernel, axis=0) #Smooting to avoid weird sampling

    new_data = interpolate.interp1d(timepoints, data, axis=0)(ref_timepoints[start_idx:stop_idx])

    idx = ref_timepoints.idx + start_idx
    return DataChunk(data=new_data, idx = idx, group=group)

# Cell
def link_sync_timepoints(frame_tp_1, frame_sig_1, frame_tp_2, frame_sig_2):
    """
    Creates timepoints between two timepoints array sampled at the same rate.
    This is usefull for the LED dome which cannot generate frames in between stimuli (due to ROM update)

    params:
        - frame_tp_1: Timepoints of the first part
        - frame_sig_1: Signals of the first part
        - frame_tp_2: Timepoints of the second part
        - frame_sig_2: Signals of the second part
    return:
        - (concatenated_frame_timepoints, concatenated_frame_signals)
    """
    assert abs(np.diff(frame_tp_1).mean() - np.diff(frame_tp_2).mean())<10, "The frame rates are different"
    assert len(frame_tp_1)==len(frame_sig_1), "The lenght of the first signals and timepoints do not match"
    assert len(frame_tp_2)==len(frame_sig_2), "The lenght of the second signals and timepoints do not match"

    n_tp = np.diff(frame_tp_1).mean()
    n_new_frames = int(round((frame_tp_2[0] - frame_tp_1[-1])/n_tp) - 1)
    new_frames = np.linspace(int(frame_tp_1[-1]+n_tp), frame_tp_2[0], n_new_frames, endpoint=False).astype(int)

    concat_frame_tp  = np.concatenate((frame_tp_1, new_frames, frame_tp_2))
    concat_frame_sig = np.concatenate((frame_sig_1, [0]*n_new_frames, frame_sig_2))

    return concat_frame_tp, concat_frame_sig

# Cell
def flip_stimulus(stim_inten, ud_inv, lr_inv):
    """
    Flip QDSpy stimuli arrays to match the up/down left/right orientation of the stimulus displayed to
    the mouse.

    params:
        - stim_inten: Stimulus matrix to flip of shape (t, color, y, x)
        - ud_inv: Up down inversion boolean (1 to make the flip, 0 for no operation)
        - lr_inv: Up down inversion boolean (1 to make the flip, 0 for no operation)

    return:
        - Flipped stimulus array
    """
    if lr_inv:
        stim_inten = np.flip(stim_inten, axis=3) # Axis 0:t 1:color 2:y 3:x
    if not ud_inv:
        #Numpy and QDSpy orientation are different.
        #This way reorientate the stimulus approriatly for display with matplotlib and potential
        #eye tracking corrections
        stim_inten = np.flip(stim_inten, axis=2)
    return stim_inten

def flip_gratings(stim_shader, ud_inv, lr_inv):
    """
    Flip gratings to match the up/down left/right orientation of the stimulus displayed to
    the mouse. A grating is encoded by an array of shape (t, 3(size, angle, speed)).
    Therefore the angles of the grating are modified to encode the "flipped" grating.

    params:
        - stim_shader: Grating matrix to flip of shape (t, 3(size, angle(degree), speed))
        - ud_inv: Up down inversion boolean (1 to make the flip, 0 for no operation)
        - lr_inv: Up down inversion boolean (1 to make the flip, 0 for no operation)

    return:
        - Flipped grating array
    """
    mask_epochs = ~np.all(stim_shader==0,axis=1)
    if lr_inv:
        stim_shader[mask_epochs,1] = (360 + (180 - stim_shader[mask_epochs,1])) % 360
    if ud_inv:
        stim_shader[mask_epochs,1] = (360 - stim_shader[mask_epochs,1]) % 360
    return stim_shader

def stim_to_dataChunk(stim_inten, stim_start_idx, reference:DataChunk) -> DataChunk:
    """
    Factory function for DataChunk of a stimulus, that squeeze the stim_inten matrix.

    params:
        - stim_inten: Stimulus matrix of shape (t, ...)
        - stim_start_idx: Starting frame index of the stimulus
        - reference: DataChunk signal reference used to determine the starting index of the stimulus

    return:
        - Datachunk of the stimulus
    """
    return DataChunk(data=np.squeeze(stim_inten), idx = (stim_start_idx + reference.idx), group="stim")

# Cell
def phy_results_dict(phy_dir):
    """
    Open the result arrays of spike sorting after manual merging with phy.

    params:
        - phy_dir: path to the phy results

    return:
        - Dictionnary of the phy arrays (amplitudes, channel_map, channel_positions, spike_clusters,
        spike_templates, spike_times, templates)
    """
    res_dict = {}
    res_dict["amplitudes"] = np.load(phy_dir+"/amplitudes.npy")
    res_dict["channel_map"] = np.load(phy_dir+"/channel_map.npy")
    res_dict["channel_positions"] = np.load(phy_dir+"/channel_positions.npy")
    res_dict["spike_clusters"] = np.load(phy_dir+"/spike_clusters.npy")
    res_dict["spike_templates"] = np.load(phy_dir+"/spike_templates.npy")
    res_dict["spike_times"] = np.load(phy_dir+"/spike_times.npy")
    res_dict["templates"] = np.load(phy_dir+"/templates.npy")
    if os.path.isfile(phy_dir+"/channel_shanks.npy"): #Newer version of phy/spyking-circus
        res_dict["channel_shanks"] = np.load(phy_dir+"/channel_shanks.npy")
        res_dict["template_ind"]   = np.load(phy_dir+"/template_ind.npy")

    return res_dict

def spike_to_dataChunk(spike_timepoints, ref_timepoints:DataChunk) -> DataChunk:
    """
    Factory function of a DataChunk for spiking count of cells from spike timepoints.

    params:
        - spike_timepoints: Dictionnary of the cells spike timepoints (list)
        - ref_timepoints: Reference DataChunk to align the newly created spike count Datachunk

    return:
        - Spike count datachunk of shape (t, n_cell)
    """
    type_cast = type(list(spike_timepoints.keys())[0])
    cell_keys = sorted(map(int,
                                    spike_timepoints.keys()))
    cell_map = dict([ (cell_key, i) for i, cell_key in enumerate(cell_keys) ])
    spike_bins = np.zeros((ref_timepoints.shape[0], len(cell_keys)))
    bins = np.concatenate((ref_timepoints[:], [(ref_timepoints[-1]*2)-ref_timepoints[-2]]))

    for i, cell in enumerate(cell_keys):
        spike_bins[:, i] = np.histogram(spike_timepoints[type_cast(cell)], bins)[0]

    datachunk = DataChunk(data=spike_bins, idx = ref_timepoints.idx, group="cell")
    datachunk.attrs["cell_map"] = cell_map
    return datachunk

# Cell
def get_calcium_stack_lenghts(folder):
    """
    Function to extract calcium stack lenghts from imageJ macro files associated to the stacks.

    params:
        - folder: path of the folder containing the IJ macros files

    return:
        - list of stack lenghts
    """
    record_lenghts = []
    pattern_nFrame = r".*number=(\d*) .*"
    for fn in glob.glob(folder+"/*.txt"):
        with open(fn) as f:
            line = f.readline()
            record_lenghts.append(int(re.findall(pattern_nFrame, line)[0]))
    return record_lenghts

def twoP_dataChunks(ref_timepoints:DataChunk, frame_timepoints, len_epochs, *args):
    """
    Factory function for two photon data.

    params:
        - ref_timepoints: Reference timepoints to create the DataChunk
        - frame_timepoints: List of frame timepoints for each sequence of two photon frame recorded.
        - len_epochs: Lenght of the recorded epochs (<= than the corresponding frame_timepoints). Int or list
        - args: matrices of all frames detected by CaImAn. (give as many as you want to synchronise)

    return:
        - tuple containing the synchronised matrices in the order it was given
    """
    assert len(args)>=1, "no matrix to be synchronised was given"
    res_l = [[] for i in range(len(args))]
    cursor = 0
    if isinstance(len_epochs, int):
        len_epochs = [len_epochs]
    # For every recording block (defined by len_epochs),
    for i, len_epoch in enumerate(len_epochs):
        start_idx = np.argmax(ref_timepoints>frame_timepoints[i][0])
        stop_idx  = np.argmax(ref_timepoints>frame_timepoints[i][len_epoch-1])
        for k, matrix in enumerate(args):
            sub_mat = matrix.T[cursor:cursor+len_epoch]

            f = interpolate.interp1d(range(len_epoch), sub_mat, axis=0)
            res_l[k].append(DataChunk(data=f(np.linspace(0,len_epoch-1,stop_idx-start_idx)),
                                           idx=start_idx,
                                           group="cell"))
        cursor += len_epoch

    return tuple(res_l)

# Cell
def img_2d_fit(shape, param_d, f):
    """
    Helper function to generate the 2D image of a fit.

    params:
        - shape: Shape of the image in (y, x).
        - param_d: Fit dictionnary.
        - f: Function used of the fit.
    """
    y_, x_ = shape
    xy = np.meshgrid(range(x_), range(y_))
    return f(xy, **param_d).reshape(y_, x_)

# Cell
def fill_nan(A):
    """
    Fill nan values with interpolation. Credits to BRYAN WOODS@StackOverflow
    """
    inds = np.arange(A.shape[0])
    good = np.where(np.isfinite(A))
    f = interpolate.interp1d(inds[good], A[good],bounds_error=False)
    B = np.where(np.isfinite(A),A,f(inds))
    return B

# Cell
def stim_inten_norm(stim_inten):
    """
    Normalize a stimulus with intensity in the 8bit range (0-255) to -1 to 1 range.
    """
    stim_inten = stim_inten.astype(float)
    stim_inten -= np.min(stim_inten)
    stim_inten -= np.max(stim_inten)/2
    stim_inten /= np.max(np.abs(stim_inten))
    return np.round(stim_inten, 5)

# Cell
def group_direction_response(stim_prop, spike_counts, n_repeat, n_cond=32):
    """
    Group the cells responses from shuffled grating stimulus repetitions. Retrieves a dictionnary
    with a key for each condition.

    params:
        - stim_prop: Grating array of shape (t, 3(size, angle, speed))
        - spike_counts: Spike counts response of the cells of shape (t, n_cell)
        - n_repeat: Number of repeat of each condition
        - n_cond: Total number of condition (speed/size condition * n_angle)

    return:
        - dictionnary of the spike counts for each condition (speed/size), with shape (n_angle, n_repeat, len, n_cell)
    """

    n_cell = spike_counts.shape[-1]
    condition_repeat = stim_prop.reshape(n_repeat*n_cond,-1,3)[:,10,:] #Take the condition for each repeat
    # We take it at the 10th frame in case of frame replacement during synchronisation
    #(the 10th should be unchanged)

    #Reshape the spike response to (n_cond, len, n_cell)
    spike_resh       = spike_counts.reshape(n_repeat*n_cond,-1,n_cell)

    angles = np.unique(condition_repeat[:,1])

    data_dict = {}
    for cond in np.unique(condition_repeat, axis=0):
        spat_freq, angle, speed = tuple(cond)
        idx_cond = np.argwhere(np.all(condition_repeat==cond, axis=1))[:,0]

        cond_key = str(spat_freq)+"@"+str(round(speed,2))
        if cond_key not in data_dict.keys():
            data_dict[cond_key] = np.empty((len(angles), len(idx_cond), *spike_resh[0].shape))

        idx_angle = np.where(angle==angles)[0][0]
        data_dict[cond_key][idx_angle] = np.array([spike_resh[idx] for idx in idx_cond])
    return data_dict

# Cell
def group_chirp_bumps(stim_inten, spike_counts, n_repeat):
    """
    Find the cells response to the OFF-ON-OFF initial parts of the chirps.

    params:
        - stim_inten: Stimulus intensity array
        - spike_counts: Spike counts array of shape (t, n_cell)
        - n_repeat: Number of repetitions of the chirp stimulus

    return:
        - Dictionnary of cells response to the different ON or OFF stimuli
    """

    repeat = stim_inten.reshape(n_repeat,-1)[0]
    spike_counts = spike_counts.reshape(n_repeat,-1,spike_counts.shape[-1])
    epoch_l = [0]
    end_l = [len(repeat)]
    i = 1
    curr = repeat[0]

    while True:
        while repeat[i]==curr:
            i+=1
        epoch_l.append(i)
        curr = repeat[i]
        if curr==repeat[i+1]:
            continue
        else:
            break

    i = len(repeat)-2
    curr = repeat[-1]

    while True:
        while repeat[i]==curr:
            i-=1
        end_l.insert(0,i)
        curr = repeat[i]
        if curr==repeat[i-1]:
            continue
        else:
            break
    slices = [slice(epoch_l[i-1],epoch_l[i]) for i in range(1,len(epoch_l))]
    slices.extend([slice(end_l[i-1],end_l[i]) for i in range(1,len(end_l))])

    res_d = {}
    for slc in slices:
        key = str(stim_inten[slc.start])+"@"+str(slc.start)
        res_d[key] = spike_counts[:,slc]

    return res_d

# Cell
def get_repeat_corrected(stim_inten, spike_counts, n_repeats=10):
    """
    Apply shifts (detected during synchro) to the chirp repetition.

    params:
        - stim_inten: Stimulus DataChunk (containing the shifts and frame replacements info)
        - spike_counts: Spike count matrix of shape (t, n_cell)
        - n_repeats: Number of repeats of the chirp

    return:
        - aligned cells response to stimulus, of shape (n_repeat, t, n_cell)
        - Number of duplicated frame per repetition.
    """
    def count_repl_in_range(fr_replaced, _range):
        return sum([repl[0] in _range for repl in fr_replaced])

    signal_shifts     = stim_inten.attrs["signal_shifts"]
    frame_replacement = stim_inten.attrs["frame_replacement"]

    spike_count_corr = spike_counts.copy()
    shift_cursor = 0
    prev_del = np.zeros((1, spike_counts.shape[1]))
    for shift, direction in signal_shifts:
        if direction=="ins":
            spike_count_corr[shift+1:] = spike_count_corr[shift:-1]
            prev_del = spike_count_corr[-1:]
        else:
            spike_count_corr[shift-1:-1] = spike_count_corr[shift:]
            spike_count_corr[-1:] = prev_del

    len_epoch = len(stim_inten)//n_repeats
    spike_counts_corrected = []
    errors_per_repeat      = []
    for i in range(n_repeats):
        errors_per_repeat.append(count_repl_in_range(frame_replacement, range(len_epoch*i, len_epoch*(i+1))))
        spike_counts_corrected.append(spike_count_corr[len_epoch*i:len_epoch*(i+1)])
    return np.array(spike_counts_corrected), np.array(errors_per_repeat)

# Cell
def removeSlowDrift(traces, fps=60, window=80, percentile=8):
    """
    Remove slow drifts from behavioral temporal traces such as locomotion speed obtained from the treadmill signal
    or pupil size obtained from the eye_tracking signal, by extracting a specified percentile within moving window from the signal.

    params:
        - traces: Behavioral temporal traces obtained from reM
        - fps: Sampling rate
        - window: Moving temporal window in seconds
        - percentile: Percentile to be extracted within moving window

    return:
        - Filtered temporal traces
    """
    smoothed = np.zeros(len(traces))
    n = round(window * fps)-1
    if n%2 == 0:
        n = n+1

    nBefore = math.floor((n-1)/2)
    nAfter = n - nBefore - 1

    for k in range(len(traces)):
        idx1 = max(np.array([0,k-nBefore]))
        idx2 = min(len(traces)-1,k+nAfter)
        tmpTraces = traces[idx1:idx2]
        smoothed[k] = np.percentile(tmpTraces, percentile)

    smoothed = savgol_filter(smoothed, n, 3)

    filteredTraces = traces - smoothed
    return filteredTraces

# Cell
def time_shift_test_corr(spike_counts, behav_signal, n_tests = 500, seed = 1):
    """
    Compute the null distribution of correlation between behavioral signal and spiking signal with a time shift test.

    params:
        - spike_counts: Array with spike counts for a specific neuron and data chunk from the reM
        - behav_signal: Array with behavioral signal for a specific neuron and data chunk from the reM
        - n_tests: number of used shifted signals to compute distribution
        - seed: seed for numpy function random.randint

    return:
        - null_dist_corr: Null distribution of correlation values
    """

    np.random.seed(seed)

    null_dist_corr=[]
    for i in range(n_tests):
        #Generate time-shifted behavioral test signal for shifts between 0.05*len(behav_signal) and len(behav_signal)
        test_behav_signal = np.roll(behav_signal, np.random.randint(len(behav_signal)*0.05, len(behav_signal)))
        # Compute Pearson's correlation with behavioral time-shifted test signal and spiking signal
        null_dist_corr.append(scipy.stats.pearsonr(test_behav_signal, spike_counts)[0])

    return null_dist_corr

# Cell
def cross_corr_with_lag(spike_counts, behav_signal, behav, conversion_factor_treadmill=None, removeslowdrift=True, fps=60, seconds=30):
    """
    Compute cross-correlation with lag between behavioral signal and spiking signal.
    Process signals, compute null distribution of the correlation with a time shift test and values .
    Return cross-correlation array, null-distribution array and values for plotting.

    params:
        - spike_counts: Array with spike counts for a specific neuron and data chunk from the reM
        - behav_signal: Array with behavioral signal for a specific neuron and data chunk from the reM
        - behav : String with name of behavioral signal to be analysed
        - conversion_factor : The value to convert the treadmill signal into cm/s
        - removeslowdrift: Boolean:
            False - doesn't remove slow drifts from the signal
            True - removes slow drifts by extracting a specified percentile within moving window from the signal
        - fps: Sampling rate
        - seconds: Window in seconds of the correlation lag

    return:
        - crosscorr: Cross-correlation with lag array between behavioral signal and spiking signal
        - corr_peak: Cross-correlation value at peak synchrony between behavioral signal and spiking signal
        - p_value_peak: P-value of the peak cross-correlation value
        - offset_peak: Temporal offset of the peak synchrony between behavioral signal and spiking signal in seconds
        - null_dist_corr: Null distribution of correlation values (output of 'utils.cross_corr_with_lag')
    """

    if behav == "treadmill":
        #Convert treadmill signal to running speed (cm/s)
        behav_signal = behav_signal * conversion_factor_treadmill
        behav_signal_filtered = gaussian_filter(abs(behav_signal), sigma=60)
    else:
        behav_signal_filtered = gaussian_filter(behav_signal, sigma=60)

    #Convolve signals with gaussian window of 1 second/60 frame
    spike_counts_filtered = gaussian_filter(spike_counts, sigma=60)

    if removeslowdrift:
        #Remove slow drifts from treadmill, pupil size and spiking signal
        spike_counts_detrend = removeSlowDrift(spike_counts_filtered, fps=60, window=100, percentile=8)
        behav_signal_detrend = removeSlowDrift(behav_signal_filtered, fps=60, window=100, percentile=8)
    else:
        spike_counts_detrend = spike_counts_filtered
        behav_signal_detrend = behav_signal_filtered

    #Get null distribution for correlation between behav_signal and spike_counts signal
    null_dist_corr = time_shift_test_corr(spike_counts_detrend, behav_signal_detrend, n_tests = 500)

    #Compute cross-correlation with lag and values to plot
    d1 = pd.Series(behav_signal_detrend)
    d2 = pd.Series(spike_counts_detrend)
    crosscorr = [d1.corr(d2.shift(lag)) for lag in range(-int(seconds*fps),int(seconds*fps+1))]
    offset_peak = np.around((np.ceil(len(crosscorr)/2)-np.argmax(abs(np.array(crosscorr))))/fps, decimals=3)
    corr_peak = np.max(abs(np.array(crosscorr)))
    p_value_peak = round((100-scipy.stats.percentileofscore(abs(np.array(null_dist_corr)), abs(corr_peak), kind='strict'))/100,2)

    return crosscorr, corr_peak, p_value_peak, offset_peak, null_dist_corr

# Cell
def get_inception_generator(imageset_folder, len_set=25, width=500, height=281):
    """
    Return a function to obtain inception loop images from their index.

    params:
        - imageset_folder: Path to the folder of the image sets
        - len_set: Number of images concatenated per set
        - width: image width
    return:
        - Function to obtain inception loop images from their index.
    """
    imageset_l = []
    paths = glob.glob(os.path.join(imageset_folder,"*.jpg"))
    paths_sorted = sorted(paths, key=lambda i: int(os.path.splitext(os.path.basename(i))[0].split("_")[-1]))

    for fn in paths_sorted: #Images accepted have the dimension (375,500)
        image = np.array(Image.open(fn))
        imageset_l.append(image)

    def image_yield(idx):
        if idx==-1:
            return np.zeros((height, width))+128
        set_idx = idx//25
        img_idx = idx%25
        return imageset_l[set_idx][:,width*img_idx:width*(img_idx+1), 1] #Returns a gray image

    return image_yield

# Cell
def group_omitted_epochs(stim_inten, spike_counts, n_fr_flash=4, n_fr_interflash=4, n_fr_isi=100):
    """
    Group the cells reponse to the different omitted stimulus epochs conditions (n_flashes)

    params:
        - stim_inten: The intensities of the omitted stimulus in shape (t)
        - spike_counts: Spikes counts of the cells in shape (t, n_cell)
        - n_fr_flash: Duration of a flash (ON flash during OFF baseline, OFF flash during ON baseline)
        - n_fr_interflash: Number of frames between two flashes (during an epoch)
        - n_fr_isi: Number of frames between two epochs
    return:
        - response_d_ON, response_d_OFF: Dictionnaries of the cells responses for different number of flashes repetions. Each contain an array of shape (n_cell, n_repeats, len_epoch+n_fr_isi).
    """
    starts_ON    = []
    stops_ON     = []
    n_flashes_ON = []

    counter   = 1
    i         = 0
    starts_ON.append(i)
    while i < len(stim_inten)-(n_fr_flash+n_fr_interflash):
        if stim_inten[i+(n_fr_flash*2+n_fr_interflash)]:
            break
        if stim_inten[i+(n_fr_flash+n_fr_interflash)]:
            counter += 1
            i+=(n_fr_flash+n_fr_interflash)
        else:
            stops_ON.append(i+(n_fr_flash+n_fr_interflash))
            n_flashes_ON.append(counter)
            counter = 1
            i += (n_fr_flash+n_fr_interflash+n_fr_isi)
            starts_ON.append(i)

    #Switching to the omitted OFF
    starts_OFF    = [starts_ON.pop()]
    stops_OFF     = []
    n_flashes_OFF = []
    while i < len(stim_inten)-(n_fr_flash+n_fr_interflash):
        if stim_inten[i+(n_fr_flash*2+n_fr_interflash)]==0:
            counter += 1
            i+=(n_fr_flash+n_fr_interflash)
        else:
            stops_OFF.append(i+(n_fr_flash+n_fr_interflash))
            n_flashes_OFF.append(counter)
            counter = 1
            i += (n_fr_flash+n_fr_interflash+n_fr_isi)
            starts_OFF.append(i)
    starts_OFF.pop()

    starts_ON     = np.array(starts_ON)
    stops_ON      = np.array(stops_ON)
    n_flashes_ON  = np.array(n_flashes_ON)
    starts_OFF    = np.array(starts_OFF)
    stops_OFF     = np.array(stops_OFF)
    n_flashes_OFF = np.array(n_flashes_OFF)

    response_d_ON, response_d_OFF = {}, {}
    for n_repeat in set(n_flashes_ON):
        where_cond = np.where(n_flashes_ON==n_repeat)[0]
        tmp        = np.array([spike_counts[start:stop+n_fr_isi] for start, stop in zip(starts_ON[where_cond],
                                                                                  stops_ON[where_cond])])
        response_d_ON[n_repeat] = np.transpose(tmp, (2, 0, 1))
    for n_repeat in set(n_flashes_OFF):
        where_cond = np.where(n_flashes_OFF==n_repeat)[0]
        tmp        = np.array([spike_counts[start:stop+n_fr_isi] for start, stop in zip(starts_OFF[where_cond],
                                                                                  stops_OFF[where_cond])])
        response_d_OFF[n_repeat] = np.transpose(tmp, (2, 0, 1))

    return response_d_ON, response_d_OFF

# Cell
def get_shank_channels(channel_positions, shank_dist_th=80):
    """
    Group the channels of a Buzsaki32 silicone probe into their shanks
    from the channel position.

    params:
        - channel_positions: List of channel positions
        - shank_dist_th: Distance between channels in X to rule if on same shank or not

    return:
        - array of grouped channel index of shape (n_shank(4), n_channel(8))
    """
    found      = np.zeros(len(channel_positions))
    shank_pos  = []
    chann_pos  = []

    while not np.all(found):
        next_idx   = np.argmin(found)
        next_pos   = channel_positions[next_idx][0] #getting the X position of the electrode
        this_shank = np.where(np.abs(channel_positions[:,0]-next_pos)<shank_dist_th)[0]
        chann_pos.append(this_shank)
        shank_pos.append(next_pos)
        found[this_shank] = 1

    shanks_idx = np.zeros((len(shank_pos), len(this_shank)), dtype=int) - 1 #Initialize with -1 in case of channel missing
    for i, order in enumerate(np.argsort(shank_pos)):
        shanks_idx[i,:len(chann_pos[order])] = chann_pos[order]
    return shanks_idx

# Cell
def format_pval(pval, significant_figures=2):
    """
    Helper function to format pvalue into string.
    """
    return '{:g}'.format(float('{:.{p}g}'.format(pval, p=significant_figures)))

# Cell
def stim_recap_df(reM):
    """
    Extract stimuli parameters (originally from the Database) to put them into a
    dataframe that will be displayed in the recapitulation plot.

    params:
        - reM: RecordMaster to extract stimuli parameters from

    return:
        - dataframe with the stimuli important informations
    """
    def parse_stim(stim_dc):
        param_d = {}
        param_d["hash"]        = stim_dc.attrs["md5"][:10] #the first 10 letters are more than enough
        param_d["n frames"]    = len(stim_dc)
        param_d["stimulus"]    = stim_dc.attrs["name"]

        if stim_dc.attrs["name"] in ["checkerboard", "fullfield_flicker", "flickering_bars", "flickering_bars_pr"]:
            param_d["frequency"] = stim_dc.attrs["refresh_rate"]
        elif stim_dc.attrs["name"] in ["chirp_am","chirp_fm","chirp_freq_epoch", "chirp_co"]:
            param_d["n ON"]      = int(stim_dc.attrs["tSteadyON_s"]*60)
            param_d["n OFF"]     = int(stim_dc.attrs["tSteadyOFF_s"]*60)
            param_d["n repeats"] = int(stim_dc.attrs["n_repeat"])
            if stim_dc.attrs["name"] in ["chirp_am","chirp_co"]:
                param_d["frequency"] = stim_dc.attrs["contrast_frequency"]
            elif stim_dc.attrs["name"]=="chirp_fm":
                param_d["frequency"] = stim_dc.attrs["max_frequency"]
            elif stim_dc.attrs["name"]=="chirp_freq_epoch":
                param_d["frequency"] = str([round(60/nfr,2) for nfr in dc.attrs["n_frame_cycle"]])
        elif stim_dc.attrs["name"] in ["fullfield_color_mix"]:
            param_d["n ON"]      = int(stim_dc.attrs["n_frame_on"])
            param_d["n OFF"]     = int(stim_dc.attrs["n_frame_off"])
            param_d["n repeats"] = int(stim_dc.attrs["n_repeat"])
        elif stim_dc.attrs["name"]=="moving_gratings":
            param_d["n repeats"]           = stim_dc.attrs["n_repeat"]
            param_d["n ON"]                = stim_dc.attrs["n_frame_on"]
            param_d["n OFF"]               = stim_dc.attrs["n_frame_off"]
            param_d["speeds"]              = stim_dc.attrs["speeds"]
            param_d["spatial frequencies"] = stim_dc.attrs["spatial_frequencies"]

        if "frame_replacement" in stim_dc.attrs:
            param_d["total drop"] = len(stim_dc.attrs["frame_replacement"])
        if "signal_shifts" in stim_dc.attrs:
            shift = 0
            for _, which_shift in stim_dc.attrs["signal_shifts"]:
                if which_shift=="ins":
                    shift += 1
                elif which_shift=="del":
                    shift -= 1
            param_d["total shift"] = shift

        return param_d

    series = []
    for seq in reM._sequences:
        for k, dc_l in seq:
            dc = dc_l[0]
            if dc.group == "stim":
                series.append(parse_stim(dc))
    df = pd.DataFrame.from_records(series, columns=["stimulus", "hash", "n frames", "n repeats",
                               "frequency", "n ON", "n OFF", "speeds", "spatial frequencies",
                              "total shift", "total drop"])
    df = df.fillna("")
    return df


# AUTOGENERATED! DO NOT EDIT! File to edit: 14_synchro.nested_stims.ipynb (unless otherwise specified).

__all__ = ['get_stim_ids', 'get_synced_file', 'get_synced_file_precompile', 'get_stim_dict', 'cluster_by_list',
           'find_stim_onsets', 'append_tread_movement', 'twoP_dataChunks_2', 'append_to_reM', 'master_loop']

# Cell
from .io import *
from .processing import *
from .extracting import *

from ..core import *
from ..utils import *
from ..database import *
from ..processing import *
from ..plotting import *

import numpy as np
import glob, os, datetime
import ast

def get_stim_ids(log, record_time, lastmodif_time):
    """Obtain the correct name of the stimulus, based on the name that appears in the QDSpy log

    This is needed if no central database of compiled stimuli is used, but instead stimuli are compiled in QDSpy
    every time before they are presented.

    params:
        - log: log object generated by get_QDSpy_logs
        - record_time: the stimulus presentation time, as calculated by subtracting stimulus duration from lastmodif_time
        - lastmodif_time: the timestamp of the .npy file (time of compilation, assumed to be only seconds before record_time)
    returns:
        - stim_ids: Correct stim_id for the stimulus presented after the current log stimulus start
    """
    stim_ids = []
    photo_early_end = []
    for i, stim in enumerate(log.stimuli):
        if stim.stop_time is None:
            stim.stop_time = log.stimuli[i + 1].start_time
        if (stim.start_time > record_time) & (stim.stop_time < lastmodif_time):
            stim_ids.append(stim)
        elif (stim.start_time > record_time):
            photo_early_end.append(stim)
    if len(stim_ids)==0:
        print("Check the timestamp of the corresponding Photodiode file. The diode seems to have stopped before the stim ended.")
        print("No matching stimulus. Taking first stim after start of photodiode. Check timestamps carefully.")
        stim_ids = [photo_early_end[0]]
    return stim_ids

def get_synced_file(stim_list_dir, stim_id):
    '''
        Find the stimulus in the stimulus list directory that is temporally closest to the
        stimulus in the log. Works based on the modification time of the stimulus (i.e.
        expects stimulus to be compiled shortly before display).

        params:
            - stim_list_dir: fullpath to stimuli, string
            - stim_id: stimulus read from QDSpy log, theonerig.synchro.extracting.Stimulus
                        object
        returns:
            - stim: filename of the stimulus that needs loading, str
    '''
    stims = {"stim_name": [], "stim_delta": []}
    for stim_list in os.listdir(stim_list_dir):
        stim_load = datetime.datetime.fromtimestamp(int(os.path.getmtime(os.path.join(stim_list_dir, stim_list))))
        stim_present = stim_id.start_time
        # If the stimulus was compiled before display calculate difference, otherwise set to max
        stim_delta = stim_present - stim_load if stim_present > stim_load else datetime.timedelta.max
        stims["stim_name"].append(stim_list)
        stims["stim_delta"].append(stim_delta)
    # Obtain the index of the compiletime closest to the stimulus display time
    closest_stim_idx = stims["stim_delta"].index(min(stims["stim_delta"]))
    stim_fn = stims["stim_name"][closest_stim_idx]
    stim_path = os.path.join(stim_list_dir, stim_fn)
    # Sanity check
    if not stim_id.name in stim_path:
        print("Compiled stimulus file not matching this name")
        print("stim_id: {}".format(stim_id.name))
        print("stimulus file: {}".format(os.path.basename(stim_path)))
        stim_path = os.path.join(stim_list_dir, os.path.basename(stim_path))
        # Clean up: Ugly 1-time fix
        if "npz" in stim_path:
            stim = np.load(stim_path)
            stim = stim['stim' + stim_id.filename[3:6]].flatten()
        else:
            stim = np.load(stim_path)
    else:
        stim = np.load(stim_path)

    # Some of the stimuli have a shape of repetition_number x stim_onset:
    if len(stim.shape) > 1:
        stim = stim.flatten()
    return stim, stim_path

def get_synced_file_precompile(stim_list_dir, stim_id):
    for stim_list in os.listdir(stim_list_dir):
        if stim_id.md5 in stim_list:
            stim_path = os.path.join(stim_list_dir, stim_list)
            stim = np.load(stim_path)
    # Some of the stimuli have a shape of repetition_number x stim_onset:
    if len(stim.shape) > 1:
        stim = stim.flatten()
    return stim, stim_path

def get_stim_dict(stim_dict_path, stim_filename):
    """Get the correct dictionary file that decodes the ints that encode the stimulus sequence.
    Applicable to nested stimuli, whose _intensity_.npy file contains only a list of integers.

    params:
        - stim_dict_path: [string], the path to the directory where all dicts are stored
        - stim_filename: [string], the name of the stimulus, make sure that stim_id from the QDSpy log and the name of the dictionary match

    """
    for dict in os.listdir(stim_dict_path):
        if stim_filename[0:6] in dict:
            with open(os.path.join(stim_dict_path, dict),'r') as inf:
                dict_from_file = []
                for line in inf:
                    dict_from_file.append(ast.literal_eval(line))
            return dict_from_file

def cluster_by_list(data, frame_timepoints, frame_signals, stim_list):
    """Assign the stimulus identity values from stim_list to the frames in data. stim_list contains only the
        sequence of stimuli. Those need to be expanded. Opposed to cluster_frame_signals and cluster_by_epochs no
        AUC operation is performed.
        params:
            - data: raw data used to compute the stimulus times
            - frame_timepoints: timepoints delimiting each frame
            - frame_signals: binary 1-D numpy array indicating if high_threshold was passed in 'detect_frames'
            - stim_list: 1-D numpy array containing the sequence of the stimuli presented
        returns:
            - frame_signals: [same size as frame_timepoints] stim_signals list containing the correct value from
                stim_list at every entry"""
    epoch_end, stim_change = find_stim_onsets(frame_signals)
    # Split into on times & values vs off times & values
    stim_ons = stim_change[0::2]
    #stim_ons_idx = stim_idx[0::2]
    stim_offs = stim_change[1::2]
    #stim_offs_idx = stim_idx[1::2]

    # Replace the frame_signal entries with the stimulus codes
    frame_signals[frame_signals == 0] = -1 # To avoid confusion with the '0' stimulus code
    for i,stim_type in enumerate(stim_list):
        if i >= len(stim_ons):
            print("WARNING: Not all stimuli of the stim_list were actually presented. Recording ended early.")
            frame_signals[stim_ons[i - 1:]] = -1
            return frame_signals, stim_ons, stim_offs, epoch_end
        frame_signals[stim_ons[i]:stim_offs[i]] = stim_type

    return frame_signals, stim_ons, stim_offs, epoch_end

def find_stim_onsets(frame_signals):
    """Get the frame timepoints during which the stimuli started and ended.
    Crop out the last two ones that are currently just an offset marker.

    params:
        - frame_signals: processed signal from the Photodiode files, containing the marker information
    returns:
        - epoch_end: the frame timepoints of the offset signal
        - stim_change: the frame timepoints of every stimulus on and offset
    """
    stim_change = np.where(frame_signals[:-1] != frame_signals[1:])[0]
    stim_change = stim_change + 1 # since I need to compare to [1:] all values are shifted by 1
    #stim_idx = frame_timepoints[stim_change]

    # QDSpy currently is set to emit a short peak at the end to indicate the end of the stimulus presentation
    # This peak needs to be ignored
    epoch_end = stim_change[-2:] # return it for future analysis
    stim_change = stim_change[:-2] #add it to the no stimulus category
    return epoch_end, stim_change

def append_tread_movement(reM, epoch_counter, treadmill_data, ref_timepoints):
    '''
    Append treadmill data to the record Master
    '''
    tread_tp = np.arange(len(treadmill_data), dtype="int")
    tread_downsamp = resample_to_timepoints(tread_tp, treadmill_data, ref_timepoints)
    reM[epoch_counter - 1]["treadmill"] = tread_downsamp
    return reM

def twoP_dataChunks_2(ref_timepoints, frame_timepoints, len_epochs, cursor, skip=None, popped_epoch=None, *args):
    """
    Factory function for two photon data.

    params:
        - ref_timepoints: Reference timepoints to create the DataChunk
        - frame_timepoints: List of frame timepoints for each sequence of two photon frame recorded.
        - len_epochs: Lenght of the recorded epochs (<= than the corresponding frame_timepoints). Int or list
        - args: matrices of all frames detected by CaImAn. (give as many as you want to synchronise)

    return:
        - tuple containing the synchronised matrices in the order it was given
    """
    assert len(args)>=1, "no matrix to be synchronised was given"
    res_l = [[] for i in range(len(args))] # A list containing 1 entry per Ca-matrix to be processed
    if cursor is None:
        cursor = 0 # counter of from where to start indexing into the Ca-matrices
    if isinstance(len_epochs, int):
        len_epochs = [len_epochs]
    if isinstance(popped_epoch, int):
        popped_epoch = [popped_epoch]
    # For every recording block (defined by len_epochs, that counts the number of frames within
    # the recording), find the index (for the ref_timepoints) of the first and last frame of
    # that block.
    for i, len_epoch in enumerate(len_epochs):
        print(i)
        if i == 0:
            set_trace()
        first_ca = frame_timepoints[i][0]
        if len_epoch > len(frame_timepoints[i]):
            last_ca = frame_timepoints[i][-1]
            print(i, "Ca data longer than sync timepoints!")
        else:
            last_ca = frame_timepoints[i][len_epoch-1]
        if isinstance(ref_timepoints, Data_Pipe):
            start_idx = np.argmax(ref_timepoints[i]['main_tp']>first_ca)
            stop_idx  = np.argmax(ref_timepoints[i]['main_tp']>last_ca)
        else:
            start_idx = np.argmax(ref_timepoints>first_ca)
            stop_idx  = np.argmax(ref_timepoints>last_ca)
        for k, matrix in enumerate(args):
            # Slice the Ca-matrix in the time dimension to the duration of the recording block
            sub_mat = matrix.T[cursor:cursor+len_epoch]
            # Generate a linear interpolation function of the Ca-matrix values over the time of the
            # recording block, axis 0 is axis of the Ca-intensity values over time
            f = interpolate.interp1d(range(len_epoch), sub_mat, axis=0)
            # stop_idx-start_idx defines the number of timepoints for which values need to be
            # interpolated (i.e resolution). The block duration is again defined by len_epoch.
            # This interpolated data will be inserted as a datachunk object starting at the
            # correct start_idx (indexing ref_timpoints)
            res_l[k].append(DataChunk(data=f(np.linspace(0,len_epoch-1,stop_idx-start_idx)),
                                           idx=start_idx,
                                           group="cell"))
        if (skip is not None) and (i+1 in [skip]):
            cursor += len_epoch + popped_epoch[0] # skip the 3000 water frames
            del popped_epoch[0]
        else:
            cursor += len_epoch # Skip indexing to the start of the next recording epoch
    return tuple(res_l)

# Cell
from .io import *
from .processing import *
from .extracting import *

from ..core import *
from ..utils import *
from ..database import *
from ..processing import *
from ..plotting import *

import numpy as np
import glob, os, datetime
def append_to_reM(epoch_counter, reM, retain_template, stim_list_dir, stim_ids, record_time,
                                photo_data, stim_code_dict, frame_tp, ref_timepoints,
                               ref_signals, tom_stim_dir):
    for stim in stim_ids:
    # Can't: Lerion didn't save it there. obtaining stimulus parameters from the database
    #     stim_params = stim_param_to_dict(stim_params_df, stim.md5)
    #     first_fr, last_fr = stim_params["first_frame_id"], stim_params["last_frame_id"]

        # One of Tom's stimuli??
        if stim.name in retain_template:
#             tom_stim_dir = os.path.join("/media/asari/Tristan/01_Code/013_Leiron_Sweep/Day_1")
            # unpacked is a tuple of (frames presented, REF and shader)
            unpacked = unpack_stim_npy(tom_stim_dir, stim.md5)
            estimate_start = get_position_estimate(stim.start_time, record_time, 10000)
            if estimate_start<0:
                print(stim, "Started before the record, passing")
                continue
            # Needs to be adjusted manually, open photo_data plots and read out first
            # and last frames. Find based on ref_tp alignment (REC) with unpacked[1] (REF)
            starting_pos = input("This stimulus needs to be synchronized. Please enter the starting ref_tp: ")
            starting_pos = int(starting_pos)
            signals = ref_signals[starting_pos:starting_pos+len(unpacked[1])]
            stim_tuple, shift_log, frame_replacement  = frame_error_correction(signals,
                                                                                unpacked,
                                                                                algo="conv")
            print("\n",stim.filename, "match at",starting_pos)
            display_match(starting_pos, unpacked[1], ref_signals, stim_tuple[1])

            # For Tom's stimuli: The first presentations are short burst encoding the stimulus
            # & the last one is a peak shorter than the rest. I'm trying to crop those out below
            # This needs to be verified later
#             epoch_end, stim_changes = find_stim_onsets(ref_signals)
#             last_frame =  epoch_end[0]

#             stim_duration = np.diff(stim_changes)
            first_frame = input("Please enter the first ref_tp of the first stimulus after the prelude code: ")
            first_frame = int(first_frame)
            last_frame = input("Please enter the last ref_tp of the first stimulus before the end-of-stimulus epoch: ")
            last_frame = int(last_frame)
            stim_tuple, shift_log, frame_replacement = chop_stim_edges(first_frame, last_frame,
                                                   stim_tuple, shift_log, frame_replacement)

            idx_stim_val = 0
            # If the photodiode data was presented in the lower right corner of the screen set '0'
            ud_inv = lr_inv = 0
#             if stim_tuple[2] is not None:
#                 idx_stim_val = 2 #If there is a shader, we take the shader instead of the intensity
            if stim.name == "moving_gratings":
                flipped_inten = flip_gratings(stim_tuple[idx_stim_val], ud_inv, lr_inv)
            else:
                flipped_inten = flip_stimulus(stim_tuple[idx_stim_val], ud_inv, lr_inv)
            stim_datachunk = stim_to_dataChunk(flipped_inten,
                                               starting_pos, #We sliced out the synchro elements of the stim
                                               ref_signals)

    #         stim_datachunk.attrs.update(stim_params) # only works if database data present
            stim_datachunk.attrs["signal_shifts"]     = shift_log
            stim_datachunk.attrs["frame_replacement"] = frame_replacement
            stim_datachunk.attrs["md5"]  = stim.md5
            stim_datachunk.attrs["name"] = stim.filename #Storing the name of the stim in the datachunck aswell

            stim_key = stim.filename
            reM[epoch_counter - 1][stim_key] = stim_datachunk


        # One of Leiron's stimuli
        else:
            # 1) Get the stimulus list from the correct .py QDSpy file
            stim_list, stim_path = get_synced_file_precompile(stim_list_dir, stim)
            # 2) Expand the stimulus values to the actually presented frames
            frame_signals_clustered, stim_ons, stim_offs, epoch_end = \
                    cluster_by_list(photo_data, ref_timepoints, ref_signals, stim_list)
            i = 0
            for onset, offset in zip(stim_ons, stim_offs):

                # 3) Are not read out from the log, but from the photodiode data in this case
                # first_fr is the first actual frame of stimulus presentation, in this case the stimulus starts immediately
                first_fr = onset
                last_fr = offset

                signals = np.array(ref_signals[first_fr:first_fr+len(
                                    frame_signals_clustered[first_fr:last_fr])])
                # Error correction compares the data from the ref_signals with the assigned 'cluster'
                # values.
                stim_tuple, shift_log, frame_replacement  = frame_error_correction(signals,
                                            [np.empty_like(signals),
                                             frame_signals_clustered[first_fr:last_fr],
                                            np.empty_like(signals)],
                                                                            algo="conv")


#                 print("\n",stim.filename, "match at",first_fr)
#                 display_match(first_fr, frame_signals_clustered[first_fr:last_fr],
#                               ref_signals, corrected=None)

                stim_datachunk = stim_to_dataChunk(frame_signals_clustered[first_fr:last_fr],
                                                   first_fr, #No syncro elements in this data
                                                   ref_signals)

                # 8) Add additional info from log file
                stim_datachunk.attrs["md5"]  = stim.md5
                stim_datachunk.attrs["signal_shifts"]     = shift_log
                stim_datachunk.attrs["frame_replacement"] = frame_replacement

                #Loading the stim dictionary and putting the correct names
                stim_idx = np.unique(stim_datachunk.data)[0]
                if stim_idx == -1:
                    stim_datachunk.attrs["name"] = "not recorded, photodiode ended early"
                else:
                    stim_datachunk.attrs["name"] = stim_code_dict[stim_idx] #Storing the name of the stim in the datachunck aswell


                # 9) Update the reM
                stim_key =  stim_datachunk.attrs["name"]
                # Based on stim_key read meaning of stim_key from dictionary
                reM[epoch_counter - 1][stim_key] = stim_datachunk
                i += 1
    return reM

# Cell
from .io import *
from .processing import *
from .extracting import *

from ..core import *
from ..utils import *
from ..database import *
from ..processing import *
from ..plotting import *

import numpy as np
import glob, os, datetime

def master_loop(log, retain_template, tom_stim_dir, sync_dir, calcium_dir, stim_list_dir,
                stim_dict_dir, stack_info_dir):
#     sync_dir = os.path.join(working_dir, "sync/Session01")
    epoch_counter = 0
    sync_files = np.sort(glob.glob(sync_dir + "/Photo*"))
    l_epochs = stack_len_extraction(stack_info_dir)
#     l_epochs[-1] = l_epochs[0] = 9033
    # Files should be numbered from 00 to 99, extract that number
    assert os.path.isdir(tom_stim_dir),      "precompiled stim dir path does not exist"
    assert os.path.isdir(sync_dir),          "sync dir path does not exist"
    assert os.path.isdir(calcium_dir),       "calcium_data dir path does not exist"
    assert os.path.isdir(stim_list_dir),     "stim_list dir path does not exist"
    assert os.path.isdir(stim_dict_dir),     "stim_dict dir path does not exist"
    assert len(sync_files) > 0,              "sync dir doesn't contain any Photodiode files"
    print("len sync_files: {}".format(len(sync_files)))
    print("len l_epochs: {}".format(len(l_epochs)))
#     assert len(sync_files) == len(l_epochs), "number of sync files and recording epochs don't align"

#     # For this particular record
    skip = [12,13]
    skipped_epoch = []
    for skips in reversed(skip):
#     skip = 12 #water sync broken
        skipped_epoch.append(l_epochs.pop(skips-1))

#     del l_epochs[skip]

    twoP_timepoints = []
    for record_file in sync_files:
        record_file_number = record_file.split('_')[-1]
        epoch_counter_current = int(record_file_number)

        if (skip is not None) and (epoch_counter_current == skip):
            print("Skipping recording #{}".format(skip))
            continue

        # Maybe list all files using glob glob instead
        if epoch_counter_current < epoch_counter:
            break
        else:
            # 0) Update file counter
            epoch_counter = epoch_counter_current
            # 1) Load the data
            photo_path = os.path.join(sync_dir,"Photodiode_" + record_file_number)
            # Should be global
#             stim_list_dir = os.path.join("/media/asari/Tristan/01_Code/013_Leiron_Sweep/Day_1")
            print("Processing: " + photo_path)
            photo_data = load_sync_raw(photo_path)
            cam_data = load_sync_raw(os.path.join(sync_dir,"CamTrigs_" + record_file_number))
            treadmill_data = load_sync_raw(os.path.join(sync_dir,"Treadmill_" +
                                                        record_file_number))
            scanning2p_data = load_sync_raw(os.path.join(sync_dir,"SyncPulses_" +
                                                         record_file_number))

#             calcium_dir = os.path.join(working_dir, "calcium_data")
            twoP_timepoints.append(detect_calcium_frames(scanning2p_data)[0]) #MUST BE APPEND TO LIST
            # 2) Get the timestamps for alignment
            lastmodif_time = datetime.datetime.fromtimestamp(int(os.path.getmtime(photo_path)))
            time_lenfile   = datetime.timedelta(0, int(len(photo_data)/10000))
            record_time = lastmodif_time-time_lenfile
            date_str = str(record_time).split(" ")[0]
            # 3) Obtain the stimuli presented during the epoch
            stim_ids = get_stim_ids(log, record_time, lastmodif_time)
            # 4) Obtain the encoding of the stimuli
            print(stim_ids[0].filename)
            if stim_ids[0].name in retain_template:
                stim_code_dict = None
            else:
                stim_code_dict = get_stim_dict(stim_dict_dir,
                                           stim_ids[0].filename)[0]
            # Thresholds and sampling rate might need to be adjusted. Check photo_data
            low_th = 0.1
            if stim_ids[0].name == "water":
                high_th = 0.6
            else:
                high_th = 1.2
            frame_tp, frame_signals = detect_frames(photo_data, low_th, high_th,
                                                    increment=int(10000/60))

            ref_timepoints, ref_signals   = extend_sync_timepoints(frame_tp, frame_signals,
                                                                   up_bound=len(photo_data))
            if epoch_counter == 1:
                reM = RecordMaster([(ref_timepoints, ref_signals)])
            else:
                reM.append(ref_timepoints, ref_signals)
            # 6) Add all stimuli to the record Master
            reM = append_to_reM(epoch_counter, reM, retain_template, stim_list_dir, stim_ids, record_time,
                                photo_data, stim_code_dict, frame_tp, ref_timepoints,
                               ref_signals, tom_stim_dir)
            # 7) Add treadmill data to record Master
            reM = append_tread_movement(reM, epoch_counter, treadmill_data, ref_timepoints)

    # 8) Add the calcium recording data
    C_matrix = np.load(os.path.join(calcium_dir,"C.npy"))
    S_matrix = np.load(os.path.join(calcium_dir,"S.npy"))
    pipe = Data_Pipe(reM, ["main_tp"])
    pipe += "main_tp"
    ca2p_C_datachunks_l = []
    ca2p_S_datachunks_l = []
    ca2p_C_datachunks, ca2p_S_datachunks = twoP_dataChunks_2(pipe,
                                                               twoP_timepoints,
                                                               l_epochs, # compare to other script, shoul dbe - 1?
                                                               0, #Grey was recorded, but excluded from C matrix generation
                                                               skip,
                                                               skipped_epoch,
                                                               C_matrix,
                                                               S_matrix)
    for i, slc in enumerate(pipe._slices):
        # Add 1) Ca-data 2) movement info to every slice (i.e. Contiguous Record)
        reM[slc[0]]["C_matrix"] = ca2p_C_datachunks[i]
        reM[slc[0]]["S_matrix"] = ca2p_S_datachunks[i]

    # 9) Determine still and movement epochs
    # Slice only the parts of the treadmill record when stimuli were presented
    pipe = Data_Pipe(reM, ["treadmill"])
    pipe += "stim"


    movement_threshold = 0.2
    for data_d, slc in zip(pipe, pipe._slices):
        treadmill_dc = data_d["treadmill"]
        tread_mean_speed = np.array(treadmill_dc).mean() # Needs type conversion

        if np.abs(tread_mean_speed) > movement_threshold:
            reM[slc[0]]["moving"] = DataChunk(np.array(treadmill_dc), idx=treadmill_dc.idx,
                                             group="data", fill=treadmill_dc.fill)
        elif np.abs(tread_mean_speed) <= movement_threshold:
            reM[slc[0]]["still"] = DataChunk(np.array(treadmill_dc), idx=treadmill_dc.idx,
                                             group="data", fill=treadmill_dc.fill)

    return reM



# AUTOGENERATED! DO NOT EDIT! File to edit: 00_core.ipynb (unless otherwise specified).

__all__ = ['DataChunk', 'ContiguousRecord', 'RecordMaster', 'Data_Pipe', 'export_record', 'import_record']

# Cell
import h5py
import json, re
import numpy as np
from collections import namedtuple
from typing import Dict, Tuple, Sequence, Union
import itertools
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

class DataChunk(np.ndarray):
    """Base brick of data. Derived from np.ndarray
    params:
        - data: The ndarray with shape (time, ...)
        - idx: Index of the start of the DataChunk in the record.
        - group: group of the DataChunk in {stim, sync, cell, data}
        - fill: Default filling value."""
    def __new__(cls, data, idx, group, fill=0):
        # See https://docs.scipy.org/doc/numpy-1.11.0/user/basics.subclassing.html#basics-subclassing
        # for explanation on subclassing numpy arrays
        obj = np.asarray(data).view(cls)
        obj.idx = idx
        obj.group = group
        obj.fill = fill

        obj.attrs = {}

        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.idx   = getattr(obj, 'idx', None)
        self.group = getattr(obj, 'group', None)
        self.fill  = getattr(obj, 'fill', 0)
        self.attrs = getattr(obj, 'attrs', {})


    @property
    def range(self):
        return range(self.idx, self.idx + len(self))

    @property
    def slice(self):
        return slice(self.idx, self.idx + len(self))

    def __str__(self):
        return ("Group: "+str(self.group)
                +"\nStarting index: "+str(self.idx)
                +"\nFilling value: "+str(self.fill)
                +"\n"+super().__str__())

    def __repr__(self):
        return "DataChunk(%s,%s,%s,%s)"%(self.shape, self.idx, self.group, self.fill)

# Cell
class ContiguousRecord():
    """Representation of a contiguous recording session to store DataChunk
    of various sources under a single time reference. DataChunk are stored
    under a name in one of the groups "sync","stim","data" and "cell".

    A name can contain multiple DataChunk if those are not overlapping in time.

    Each ContiguousRecord contains in the group "sync" two master DataChunk,
    one for signals to be recorded across acquisition device to syncronize them,
    one for timepoints of these signals for the main device and are called
    respectively "signals" and "main_tp".

    """
    MAIN_TP = "main_tp"
    SIGNALS = "signals"
    def __init__(self, length:int, signals:DataChunk, main_tp:DataChunk, frame_rate:int=60 ):
        """Instanciate a ContiguousRecord.

        Parameters:
            - length (int): Number of bins of this record
            - signals (DataChunk): Signals for this record
            - main_tp (DataChunk): Timepoints of the signals for the main device
            - frame_rate(int): Frame rate in Hz
        """
        self.length = length
        self._frame_time = 1/frame_rate
        self._data_dict = {}

        self[self.SIGNALS] = signals
        self[self.MAIN_TP] = main_tp

        self._slice = slice(0,self.length,1)

    def dataset_intersect(self, existing_datachunk:list, new_datachunk:DataChunk):
        """Check for timepoint intersections of two DataChunks"""
        range_new = set(new_datachunk.range)
        intersect = False
        for range_existing in existing_datachunk:
            intersect |= len(range_new.intersection(range_existing.range)) > 0
        return intersect

    def keys(self):
        """Retrieves the existing keyys inside this ContiguousRecord"""
        return self._data_dict.keys()

    def get_slice(self, datachunk_name:str) -> list:
        """Returns the slices of the DataChunk corresponding to the given key"""
        if datachunk_name in self._data_dict.keys():
            return [chunk.slice for chunk in self._data_dict[datachunk_name]]
        else:
            return []

    def set_slice(self, slice_):
        """Set the slice to restrict the size of the DataChunk returned"""
        if slice_ is None:
            self._slice = slice(0,self.length,1)
        else:
            start, stop, step = slice_.start, slice_.stop, slice_.step
            if start is None:
                start = 0
            if stop is None:
                stop = self.length
            if step is None:
                step = 1
            if step!=1:
                print("Step in slice is currently not supported.")
                self._slice = slice(0,self.length,1)
            else:
                self._slice = slice(start,stop,step)


    def get_names_group(self, group_name:str) -> list:
        names = []
        for key, dChunk_l in self._data_dict.items():
            if dChunk_l[0].group == group_name:
                names.append(key)
        return names

    def to_s(self, n_frame):
        return round(self._frame_time*n_frame,2)

    def to_time_str(self, n_frame):
        s = int(self.to_s(n_frame))
        m, s = s//60, str(s%60)
        h, m = str(m//60), str(m%60)
        return "{0}:{1}:{2}".format('0'*(2-len(h))+h, '0'*(2-len(m))+m, '0'*(2-len(s))+s)

    def __len__(self):
        return self.length

    def __setitem__(self, key, value:DataChunk):
        if isinstance(key, str):
            if key not in self._data_dict.keys():
                self._data_dict[key] = []

            if not self.dataset_intersect(self._data_dict[key], value):
                self._data_dict[key].append(value)
            else:
                raise ValueError("Data with the same name already exists and intersect with the one provided")
        else:
            raise KeyError("Cannot set data with an integer index, it needs a name")

    def __getitem__(self, key):
        if isinstance(key, str):
            l_datachunk = self._data_dict[key]
            fill_value  = l_datachunk[0].fill
            shape       = l_datachunk[0].shape

            full_sequence = DataChunk(np.zeros((len(range(*self._slice.indices(self.length))), *shape[1:]),
                                               dtype=l_datachunk[0].dtype)+fill_value,
                                      self._slice.start if self._slice.start is not None else 0,
                                      fill_value)
            for datachunk in l_datachunk:
                dc_slice = datachunk.slice
                if dc_slice.start>=self._slice.stop or dc_slice.stop<=self._slice.start:
                    continue

                start = max(dc_slice.start, self._slice.start) #flooring to the maximum of both start
                stop  = min(dc_slice.stop, self._slice.stop) # and capping to the min of both end

                new_dc_slice  = slice(start-datachunk.idx, stop-datachunk.idx)
                res_start = self._slice.start if self._slice.start is not None else 0
                res_slice = slice(start-res_start, stop-res_start)
                full_sequence[res_slice] = datachunk.data[new_dc_slice]
                full_sequence.attrs.update(datachunk.attrs)

            return full_sequence

    def __iter__(self):
        groups = {"sync":[],"stim":[],"data":[],"cell":[]}
        for key, dChunk_l in self._data_dict.items():
            groups[dChunk_l[0].group].append((dChunk_l[0].idx, key))

        self._iter_order = []
        for group_name in ["sync","stim","data","cell"]:
            sorted_ = sorted(groups[group_name], key=lambda e:(e[0],))
            self._iter_order.extend([key for _, key in sorted_])
        self._n = 0

        return self

    def __next__(self):
        if self._n < len(self._iter_order):
            key = self._iter_order[self._n]
            dChunk_l = self._data_dict[key]
            self._n += 1
            return (key, dChunk_l)
        else:
            raise StopIteration

    def __delitem__(self, key):
        del self._data_dict[key]

    def __str__(self):
        res = "ContiguousRecord:\n"
        for k,v in self._data_dict.items():
            res += k+" : "+" ".join([str(dc.shape) for dc in v]) +"\n"
        return res

    def __repr__(self):
        return self._data_dict.__repr__()

    def __delete__(self, instance):
        for k, v in self._data_dict.items():
            del v
            self._data_dict[k] = None

# Cell
class RecordMaster(list):
    """
    The RecordMaster class is the top level object managing all
    timeseries. It uses a list of ContiguousRecord to represent
    possible discontinuted data records.

    The main aim of the RecordMaster is to store the various data
    stream of an experiment under a unique time reference, to ease
    the processing of the data.

    A RecordMaster is created by providing to it a list of (timepoints, signals) arrays of
    identical lenght. This serves as reference to align other DataChunks with the signals
    at the given timepoints. Multiple tuples represent uncontiguous sequences of data in
    a same record. The timepoints must be evenly spaced (regular).

    params:
        - reference_data_list: list of (timepoints, signals) arrays.
        - frame_rate: Frame rate in Hz, or list of frame rates matching len(reference_data_list)
    """

    def __init__(self, reference_data_list: Sequence[Tuple[DataChunk, DataChunk]], frame_rate=60):

        if not hasattr(frame_rate, '__iter__'):
            frame_rate = [frame_rate]*len(reference_data_list)

        self._sep_size   = 1000 #Used for the plotting of multiple sequences
        self._sequences = []
        for (ref_timepoints, ref_signals), fr in zip(reference_data_list, frame_rate):
            cs = ContiguousRecord(len(ref_timepoints), ref_signals, ref_timepoints, fr)
            self._sequences.append(cs)

    def set_datachunk(self, dc:DataChunk, name:str, sequence_idx=0):
        """Set the given DataChunk dc for the sequence at sequence_idx under name."""
        self._sequences[sequence_idx][name] = dc

    def append(self, ref_timepoints:DataChunk, ref_signals:DataChunk, frame_rate=60):
        cs = ContiguousRecord(len(ref_timepoints), ref_signals, ref_timepoints, frame_rate)
        self._sequences.append(cs)

    def insert(self, idx:int, ref_timepoints:DataChunk, ref_signals:DataChunk, frame_rate=60):
        cs = ContiguousRecord(len(ref_timepoints), ref_signals, ref_timepoints, frame_rate)
        self._sequences.insert(idx, cs)

    def keys(self):
        keys = []
        for seq in self._sequences:
            keys.extend(list(seq.keys()))
        return set(keys)

    def __setitem__(self, key, value:DataChunk):
        """Setting an item directly to the record_master place it in the first sequence"""
        if isinstance(key, str):
            self._sequences[0][key] = value

    def __getitem__(self, key):
        if isinstance(key, (int, np.integer)):
            return self._sequences[key]
        elif isinstance(key, str):
            #We want all the data of that name
            res = []
            for seq in self._sequences:
                res.append(seq[key])
            return res

        raise TypeError("Indexing not understood")

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self):
            res = self._sequences[self._n]
            self._n += 1
            return res
        else:
            raise StopIteration

    def __len__(self):
        return len(self._sequences)

    def plot(self, ax=None, show_time=True, sort_by_name=False):
        colors = {"sync":"cornflowerblue", "stim":"orange", "data":"yellowgreen", "cell":"plum"}
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))
        ax.invert_yaxis()
        ax.xaxis.set_visible(False)

        #A first iteration through the ContiguousRecords to have all the DataChunk for the sorting
        groups = {"sync":[],"stim":[],"data":[],"cell":[]}
        frame_times = []
        for seq in self._sequences:
            frame_times.append(seq._frame_time)
            for name, dChunk_l in seq:
                dc = dChunk_l[0]
                if name not in groups[dc.group]:
                    groups[dc.group].append(name)

        all_names = []
        y_pos_dict = {}
        y_count    = 0
        for group_name in ["sync","stim","data","cell"]:
            if sort_by_name:
                groups[group_name].sort()
            all_names += groups[group_name]
            for dc_name in groups[group_name]:
                y_pos_dict[dc_name] = y_count
                y_count+=1

        frametime_ratios = [fr_t/max(frame_times) for fr_t in frame_times]
        cursor = 0
        for i, seq in enumerate(self._sequences):
            fr_ratio = frametime_ratios[i]
            for name, dChunk_l in seq:
                y_pos = y_pos_dict[name]
                for dChunk in dChunk_l:
                    pos = dChunk.idx + cursor
                    ax.barh(y_pos, len(dChunk)*fr_ratio, left=pos*fr_ratio, height=0.8, color=colors[dChunk.group], label=dChunk.group)
                    if show_time:
                        x = pos + len(dChunk)/2
                        text = "{0} -> {1} ".format(seq.to_time_str(dChunk.idx), seq.to_time_str(dChunk.idx+len(dChunk)))
                        ax.text(x*fr_ratio, y_pos, text, ha='center', va='center')
            cursor += (len(seq))*fr_ratio + self._sep_size

        legend_elements = [Patch(facecolor=colors["sync"],label='Synchro'),
                           Patch(facecolor=colors["data"],label='Data'),
                           Patch(facecolor=colors["stim"],label='Stimulus'),
                           Patch(facecolor=colors["cell"],label='Cell'),]
        ax.legend(handles=legend_elements, ncol=5, bbox_to_anchor=(0, 1), loc='lower left', fontsize='small')
        ax.set_yticks(range(len(all_names)))
        ax.set_yticklabels(all_names)
        ax.set_xlim(-100,cursor)

    def __str__(self):
        return "["+",\n".join([repr(seq) for seq in self._sequences])+"]"

    def __repr__(self):
        return "["+", ".join([repr(seq) for seq in self._sequences])+"]"

    def __delete__(self, instance):
        for seq in self._sequences:
            del seq

# Cell
class Data_Pipe():
    """
    A Data_Pipe is used to query data from a RecordMaster. By adding/substracting portions
    of the record using DataChunk names, it creates a mask over which specified DataChunks
    are retrivied as a dictionary from the Data_Pipe.

    params:
        - record_master: the RecordMaster from which to retrieve data
        - data_names: Name, or list of names of the DataChunk to retrieve once the masking process is done.
        - target_names: Alias, or list of alias to give to the DataChunk in the retrieved dictionnary.
        - cast_to_np: boolean to cast the piped data into a numpy array. Can be set directly too in self.cast_to_np
    """
    def __init__(self, record_master:RecordMaster, data_names:Union[str,list], target_names:Union[str,list]=None, cast_to_np=False):
        self.record_master = record_master
        if isinstance(data_names, str):
            data_names = [data_names]

        if isinstance(target_names, str):
            target_names = [target_names]

        if target_names is None:
            target_names = data_names
        elif len(data_names) != len(target_names):
            raise Exception("data_names and target_names length must match")

        self.target_names = target_names
        self.data_names   = data_names
        self._masks       = [np.zeros(len(seq), dtype=bool) for seq in record_master]
        self._slices      = []

        self.cast_to_np   = cast_to_np

    def plot(self, newfig=False):
        """
        Add the mask of the Pipe to a RecordMaster plot.
        Usage:
            reM.plot()
            pipe.plot()
        """
        factor = -1
        if newfig:
            plt.figure()
            factor = 1
        cursor = 0
        frame_times      = [seq._frame_time for seq in self.record_master._sequences]
        frametime_ratios = [fr_t/max(frame_times) for fr_t in frame_times]
        for i, seq in enumerate(self.record_master):
            fr = frametime_ratios[i]
            x  = np.linspace(cursor, cursor+len(seq)*fr, len(seq), endpoint=False)
            plt.plot(x, self._masks[i]*factor-1, c='tab:blue')
            cursor += len(seq)*fr + self.record_master._sep_size

    def copy(self):
        """
        Duplicate the Pipe and its mask, and return the new pipe.
        """
        new_pipe =  Data_Pipe(record_master=self.record_master,
                         data_names=self.data_names,
                         target_names=self.target_names)
        new_pipe._masks = np.copy(self._masks)
        new_pipe._slices = self._slices.copy()
        return new_pipe

    def _get_dchunk_names(self, names):
        if isinstance(names, str):
            names = [names]
        dchunk_name = []
        for seq in self.record_master:
            for name in names:
                if name in ["sync", "cell", "data", "stim"]:
                    dchunk_name.extend(seq.get_names_group(name))
                else:
                    dchunk_name.append(name)
        return list(set(dchunk_name))

    def _intersect_names(self):
        for i, seq in enumerate(self.record_master):
            for name in self.data_names:
                if name not in seq.keys():
                    self._masks[i] = np.zeros(len(seq), dtype=bool)
                    break

    def _update_slices(self):
#         self._intersect_names() #Always intersect the names we wanna retrieve ? Might be not needed (even cause bug)
        self._slices = []
        #Iterating the list of mask (one per seq of the record_master)
        for j, mask in enumerate(self._masks):
            seq = self.record_master[j]
            mask = np.concatenate(([0],mask,[0])) #Putting zeros on the side in case the limits would be ones
            for start, stop in np.where(mask[1:]-mask[:-1])[0].reshape(-1,2):#Split the mask where 0 separates the 1
                self._slices.append((j, slice(start,stop)))

    def __ior__(self, names:Union[str, list]):
        return self.__iadd__(names)
    def __or__(self, names:Union[str, list]):
        return self.copy().__ior__(names)

    def __iand__(self, names:Union[str, list]):
        dchunk_name = self._get_dchunk_names(names)
        for i, seq in enumerate(self.record_master):
            new_mask = np.zeros(len(seq), dtype=bool)
            for name in dchunk_name:
                for slice_ in seq.get_slice(name):
                    new_mask[slice_] = 1
            self._masks[i] &= new_mask
        self._update_slices()
        return self
    def __and__(self, names:Union[str, list]):
        return self.copy().__iand__(names)

    def __ixor__(self, names:Union[str, list]):
        dchunk_name = self._get_dchunk_names(names)
        for i, seq in enumerate(self.record_master):
            new_mask = np.zeros(len(seq), dtype=bool)
            for name in dchunk_name:
                for slice_ in seq.get_slice(name):
                    new_mask[slice_] = 1
            self._masks[i] ^= new_mask
        self._update_slices()
        return self
    def __xor__(self, names:Union[str, list]):
        return self.copy().__ixor__(names)

    def __iadd__(self, names:Union[str, list]):
        dchunk_name = self._get_dchunk_names(names)
        for i, seq in enumerate(self.record_master):
            for name in dchunk_name:
                for slice_ in seq.get_slice(name):
                    self._masks[i][slice_] = 1
        self._update_slices()
        return self
    def __add__(self, names:Union[str, list]):
        return self.copy().__iadd__(names)

    def __isub__(self, names:Union[str, list]):
        dchunk_name = self._get_dchunk_names(names)
        for i, seq in enumerate(self.record_master):
            for name in dchunk_name:
                for slice_ in seq.get_slice(name):
                    self._masks[i][slice_] = 0
        self._update_slices()
        return self
    def __sub__(self, names:Union[str, list]):
        return self.copy().__isub__(names)

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self):
            res = {}
            seq_idx, _slice = self._slices[self._n]
            self.record_master[seq_idx].set_slice(_slice)
            for i, name in enumerate(self.data_names):
                if self.cast_to_np:
                    res[self.target_names[i]] = np.array(self.record_master[seq_idx][name])
                else:
                    res[self.target_names[i]] = self.record_master[seq_idx][name]
            self.record_master[seq_idx].set_slice(None)
            self._n += 1
            return res
        else:
            raise StopIteration

    def __len__(self):
        return len(self._slices)

    def __getitem__(self, key):
        if isinstance(key, (int, np.integer)):
            seq_idx, _slice = self._slices[key]
            self.record_master[seq_idx].set_slice(_slice)
            res = {}
            for i, name in enumerate(self.data_names):
                if self.cast_to_np:
                    res[self.target_names[i]] = np.array(self.record_master[seq_idx][name])
                else:
                    res[self.target_names[i]] = self.record_master[seq_idx][name]
            self.record_master[seq_idx].set_slice(None)
            return res
        elif isinstance(key, slice):
            l_res = []
            for seq_idx, _slice in self._slices[key]:
                res = {}
                for i, name in enumerate(self.data_names):
                    if self.cast_to_np:
                        res[self.target_names[i]] = np.array(self.record_master[seq_idx][name][_slice])
                    else:
                        res[self.target_names[i]] = self.record_master[seq_idx][name][_slice]

                l_res.append(res)
            return l_res
        else:
            raise IndexError ("only integers and slices (`:`) are valid indices")

    def __str__(self):
        return "(datachunks, targets, slices), "+self.__repr__()

    def __repr__(self):
        return "Pipe(%s)"%(repr(self.data_names)+", "+repr(self.target_names)+", "+repr(self._slices))

# Cell
def export_record(path, record_master):
    """Export a Record_Master object to an h5 file, readable outside of this library.

    params:
        - path: path of the file to be saved
        - record_master: RecordMaster to save
    """
    print("Exporting the record master")
    with h5py.File(path, mode="w") as h5_f:
        fr = None
        if hasattr(record_master, '_frame_time'):
            fr = record_master._frame_time #_frame_time was moved to Contigous_Record
        h5_f.attrs["_sep_size"]   = record_master._sep_size
        for i, contig in enumerate(record_master):
            #create contig
            print("Contiguous sequence",i)
            cntig_ref = h5_f.create_group(str(i))
            cntig_ref.attrs["length"] = contig.length
            if fr is not None:
                cntig_ref.attrs["_frame_time"] = fr
            else:
                cntig_ref.attrs["_frame_time"] = contig._frame_time
            for key, dc_list in contig._data_dict.items():
                #create datastream
                print("...Entering stream",key)
                stream_ref = cntig_ref.create_group(key)
                for datachunk in dc_list:
                    print("......",str(datachunk.idx)+"->"+str(datachunk.idx+len(datachunk)))
                    dset = stream_ref.create_dataset(str(datachunk.idx), data=np.array(datachunk), compression="gzip", compression_opts=4)
                    ndarray_ref = stream_ref.create_group("__ndarray_"+str(datachunk.idx))
                    for attr_k, attr_v in datachunk.attrs.items():
                        if isinstance(attr_v, (np.ndarray,)):
                            ndarray_ref.create_dataset(attr_k, data=attr_v, compression="gzip", compression_opts=4)
                        else:
                            dset.attrs[attr_k] = json.dumps(attr_v)
                    dset.attrs["__fill"] = datachunk.fill
                    dset.attrs["__group"] = datachunk.group
    print()

def import_record(path):
    """Import a Record_Master from an h5 file saved by the export_record function of this library.

    params:
        - path: path of the RecordMaster to import
    """
    print("Importing the record master")
    with h5py.File(path, mode="r") as h5_f:
        record_master    = None
        frame_rate = None
        reg         = re.compile("frame_time")
        fr_time_key = list(filter(reg.search, h5_f.attrs.keys()))
        if len(fr_time_key)==1:
            frame_rate = round(1/h5_f.attrs[fr_time_key[0]])
        keys = sorted(h5_f.keys(), key=int)
        for j, key_contig in enumerate(keys):
            ref_contig = h5_f[key_contig]
            stream_d   = {}
            keys = sorted(h5_f.keys(), key=int)
            for i, key_dstream in enumerate(ref_contig.keys()):
                ref_dstream = ref_contig[key_dstream]
                dchunk_l = []
                for key_dc in ref_dstream.keys():
                    if key_dc.startswith("__ndarray_"): #skipping array attributes for later
                        continue
                    data = ref_dstream[key_dc]
                    idx  = int(key_dc)
                    attrs = {}
                    for k,v in data.attrs.items():
                        if k not in  ["__fill", "__group"]:
                            attrs[k] = json.loads(v)
                        elif k == "__fill":
                            fill = v
                        elif k == "__group":
                            group = v
                    if "__ndarray_"+str(idx) in ref_dstream.keys():
                        # if condition for backward support
                        for k,v in ref_dstream["__ndarray_"+str(idx)].items():
                            attrs[k] = v[:]
                    dchunk = DataChunk(data=data[:], idx=idx, group=group)
                    dchunk.attrs = attrs
                    dchunk_l.append(dchunk)

                stream_d[key_dstream] = dchunk_l
            if len(fr_time_key)==0:
                frame_rate = round(1/ref_contig.attrs["_frame_time"])
            if record_master is None:
                record_master = RecordMaster([(stream_d["main_tp"][0],stream_d["signals"][0])], frame_rate=frame_rate)
            else:
                record_master.append(stream_d["main_tp"][0],stream_d["signals"][0], frame_rate=frame_rate)
            for kstream, vstream in stream_d.items():
                for k, dc in enumerate(vstream):
                    if kstream in ["main_tp", "signals"] and k==0:
                        continue
                    record_master.set_datachunk(dc, name=kstream, sequence_idx=j)
    print()
    return record_master


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"DataChunk": "00_core.ipynb",
         "ContiguousRecord": "00_core.ipynb",
         "RecordMaster": "00_core.ipynb",
         "Data_Pipe": "00_core.ipynb",
         "export_record": "00_core.ipynb",
         "import_record": "00_core.ipynb",
         "extend_sync_timepoints": "01_utils.ipynb",
         "align_sync_timepoints": "01_utils.ipynb",
         "resample_to_timepoints": "01_utils.ipynb",
         "link_sync_timepoints": "01_utils.ipynb",
         "flip_stimulus": "01_utils.ipynb",
         "flip_gratings": "01_utils.ipynb",
         "stim_to_dataChunk": "01_utils.ipynb",
         "phy_results_dict": "01_utils.ipynb",
         "spike_to_dataChunk": "01_utils.ipynb",
         "get_calcium_stack_lenghts": "01_utils.ipynb",
         "twoP_dataChunks": "01_utils.ipynb",
         "img_2d_fit": "01_utils.ipynb",
         "fill_nan": "01_utils.ipynb",
         "stim_inten_norm": "01_utils.ipynb",
         "group_direction_response": "01_utils.ipynb",
         "group_chirp_bumps": "01_utils.ipynb",
         "get_repeat_corrected": "01_utils.ipynb",
         "removeSlowDrift": "01_utils.ipynb",
         "time_shift_test_corr": "01_utils.ipynb",
         "cross_corr_with_lag": "01_utils.ipynb",
         "get_inception_generator": "01_utils.ipynb",
         "group_omitted_epochs": "01_utils.ipynb",
         "get_shank_channels": "01_utils.ipynb",
         "format_pval": "01_utils.ipynb",
         "stim_recap_df": "01_utils.ipynb",
         "eyetrack_stim_inten": "02_processing.ipynb",
         "saccade_distances": "02_processing.ipynb",
         "smooth_eye_position": "02_processing.ipynb",
         "process_sta_batch": "02_processing.ipynb",
         "staEst_fromBins": "02_processing.ipynb",
         "process_sta_batch_large": "02_processing.ipynb",
         "cross_correlation": "02_processing.ipynb",
         "corrcoef": "02_processing.ipynb",
         "flatten_corrcoef": "02_processing.ipynb",
         "stimulus_ensemble": "02_processing.ipynb",
         "process_nonlinearity": "02_processing.ipynb",
         "activity_histogram": "02_processing.ipynb",
         "cross_distances": "02_processing.ipynb",
         "cross_distances_sta": "02_processing.ipynb",
         "paired_distances": "02_processing.ipynb",
         "paired_distances_sta": "02_processing.ipynb",
         "direction_selectivity": "02_processing.ipynb",
         "wave_direction_selectivity": "02_processing.ipynb",
         "peri_saccadic_response": "02_processing.ipynb",
         "sigmoid": "03_modelling.ipynb",
         "gaussian": "03_modelling.ipynb",
         "sum_of_gaussian": "03_modelling.ipynb",
         "diff_of_gaussian": "03_modelling.ipynb",
         "gaussian_2D": "03_modelling.ipynb",
         "sum_of_2D_gaussian": "03_modelling.ipynb",
         "exponential_decay": "03_modelling.ipynb",
         "sin_exponent": "03_modelling.ipynb",
         "sinexp_sigm": "03_modelling.ipynb",
         "fit_sigmoid": "03_modelling.ipynb",
         "fit_spatial_sta": "03_modelling.ipynb",
         "fit_temporal_sta": "03_modelling.ipynb",
         "fit_chirp_am": "03_modelling.ipynb",
         "fit_chirp_freq_epoch": "03_modelling.ipynb",
         "fit_transiency": "03_modelling.ipynb",
         "repetition_quality_index": "03_modelling.ipynb",
         "onoff_transient_index": "03_modelling.ipynb",
         "DEFAULT_COLORS": "04_plotting.ipynb",
         "plot_2d_fit": "04_plotting.ipynb",
         "plot_tSTA_fit": "04_plotting.ipynb",
         "plot_chirpam_fit": "04_plotting.ipynb",
         "plot_chirp_freq_epoch_fit": "04_plotting.ipynb",
         "plot_transient_fit": "04_plotting.ipynb",
         "plot_nonlinearity_fit": "04_plotting.ipynb",
         "plot_ds_wheel": "04_plotting.ipynb",
         "plot_ds_wave_wheel": "04_plotting.ipynb",
         "plot_fl_bars": "04_plotting.ipynb",
         "plot_t_sta": "04_plotting.ipynb",
         "plot_chirp": "04_plotting.ipynb",
         "plot_spike_template": "04_plotting.ipynb",
         "plot_spike_template_kilo": "04_plotting.ipynb",
         "plot_spike_template_MEA": "04_plotting.ipynb",
         "plot_autocorrelogram": "04_plotting.ipynb",
         "plot_spike_amplitudes": "04_plotting.ipynb",
         "plot_spike_amplitudes_kilo": "04_plotting.ipynb",
         "plot_calcium_trace": "04_plotting.ipynb",
         "plot_stim_epochs_to_ephy": "04_plotting.ipynb",
         "plot_stim_epochs_to_calcium": "04_plotting.ipynb",
         "plot_cell_spatial": "04_plotting.ipynb",
         "plot_stim_recap_table": "04_plotting.ipynb",
         "plot_composed_A_masks": "04_plotting.ipynb",
         "plot_sta_positions": "04_plotting.ipynb",
         "plot_2d_sta": "04_plotting.ipynb",
         "plot_dome_flat": "04_plotting.ipynb",
         "plot_dome_checker": "04_plotting.ipynb",
         "plot_omitted_response": "04_plotting.ipynb",
         "plot_sta_pixelcorr": "04_plotting.ipynb",
         "plot_svd": "04_plotting.ipynb",
         "plot_nonlin": "04_plotting.ipynb",
         "plot_crosscorr_spikes_behav": "04_plotting.ipynb",
         "configure_pyplot_recap": "04_plotting.ipynb",
         "plot_recap_vivo_ephy": "04_plotting.ipynb",
         "plot_recap_vivo_ephy_kilo": "04_plotting.ipynb",
         "plot_recap_vivo_calcium": "04_plotting.ipynb",
         "plot_recap_vitro_ephy": "04_plotting.ipynb",
         "plot_recap_vivo_ephy_dome": "04_plotting.ipynb",
         "plot_recap_vivo_ephy_corr_behav": "04_plotting.ipynb",
         "plot_recap_wholeField": "04_plotting.ipynb",
         "plot_recap_wholeField_dome": "04_plotting.ipynb",
         "plot_recap_wholeField_vitroHiroki": "04_plotting.ipynb",
         "get_db_engine": "05_database.ipynb",
         "prompt_credentials": "05_database.ipynb",
         "get_record_essentials": "05_database.ipynb",
         "get_stim_params": "05_database.ipynb",
         "get_table": "05_database.ipynb",
         "stim_param_to_dict": "05_database.ipynb",
         "interpolate_screen_pos": "06_eyetrack.ipynb",
         "interpolate_checker_pos": "06_eyetrack.ipynb",
         "split_eye_events": "06_eyetrack.ipynb",
         "get_spherical_map": "06_eyetrack.ipynb",
         "apply_spherical_map": "06_eyetrack.ipynb",
         "atoi": "10_synchro.io.ipynb",
         "natural_keys": "10_synchro.io.ipynb",
         "filter_per_extension": "10_synchro.io.ipynb",
         "print_and_log": "10_synchro.io.ipynb",
         "print_info": "10_synchro.io.ipynb",
         "print_error": "10_synchro.io.ipynb",
         "get_offset": "10_synchro.io.ipynb",
         "logger": "10_synchro.io.ipynb",
         "DataFile": "10_synchro.io.ipynb",
         "read_header": "10_synchro.io.ipynb",
         "get_bytes_per_data_block": "10_synchro.io.ipynb",
         "read_qstring": "10_synchro.io.ipynb",
         "RHDFile": "10_synchro.io.ipynb",
         "H5File": "10_synchro.io.ipynb",
         "RawBinaryFile": "10_synchro.io.ipynb",
         "NumpyFile": "10_synchro.io.ipynb",
         "load_all_data": "10_synchro.io.ipynb",
         "load_all_data_adc": "10_synchro.io.ipynb",
         "load_all_data_dig_in": "10_synchro.io.ipynb",
         "load_all_data_both": "10_synchro.io.ipynb",
         "export_adc_raw": "10_synchro.io.ipynb",
         "export_dig_in_raw": "10_synchro.io.ipynb",
         "export_raw": "10_synchro.io.ipynb",
         "export_both_raw": "10_synchro.io.ipynb",
         "load_adc_raw": "10_synchro.io.ipynb",
         "load_digin_raw": "10_synchro.io.ipynb",
         "load_sync_raw": "10_synchro.io.ipynb",
         "get_QDSpy_logs": "11_synchro.extracting.ipynb",
         "QDSpy_log": "11_synchro.extracting.ipynb",
         "Stimulus": "11_synchro.extracting.ipynb",
         "get_synced_file": "14_synchro.nested_stims.ipynb",
         "unpack_stim_npy": "11_synchro.extracting.ipynb",
         "extract_spyking_circus_results": "11_synchro.extracting.ipynb",
         "extract_best_pupil": "11_synchro.extracting.ipynb",
         "stack_len_extraction": "11_synchro.extracting.ipynb",
         "get_thresholds": "12_synchro.processing.ipynb",
         "get_first_high": "12_synchro.processing.ipynb",
         "detect_frames": "12_synchro.processing.ipynb",
         "reverse_detection": "12_synchro.processing.ipynb",
         "extend_timepoints": "12_synchro.processing.ipynb",
         "error_check": "12_synchro.processing.ipynb",
         "cluster_frame_signals": "12_synchro.processing.ipynb",
         "cluster_by_epochs": "12_synchro.processing.ipynb",
         "cluster_by_list": "14_synchro.nested_stims.ipynb",
         "parse_time": "12_synchro.processing.ipynb",
         "get_position_estimate": "12_synchro.processing.ipynb",
         "match_starting_position": "12_synchro.processing.ipynb",
         "display_match": "12_synchro.processing.ipynb",
         "frame_error_correction": "12_synchro.processing.ipynb",
         "error_frame_matches": "12_synchro.processing.ipynb",
         "apply_shifts": "12_synchro.processing.ipynb",
         "shift_detection_conv": "12_synchro.processing.ipynb",
         "shift_detection_NW": "12_synchro.processing.ipynb",
         "chop_stim_edges": "12_synchro.processing.ipynb",
         "detect_calcium_frames": "12_synchro.processing.ipynb",
         "get_dome_positions": "13_leddome.ipynb",
         "as_cartesian": "13_leddome.ipynb",
         "as_spherical": "13_leddome.ipynb",
         "angular_distance": "13_leddome.ipynb",
         "build_wave_stimulus_array": "13_leddome.ipynb",
         "Quaternion": "13_leddome.ipynb",
         "get_waves_relative_position": "13_leddome.ipynb",
         "get_led_relative_position": "13_leddome.ipynb",
         "get_stim_ids": "14_synchro.nested_stims.ipynb",
         "get_synced_file_precompile": "14_synchro.nested_stims.ipynb",
         "get_stim_dict": "14_synchro.nested_stims.ipynb",
         "find_stim_onsets": "14_synchro.nested_stims.ipynb",
         "append_tread_movement": "14_synchro.nested_stims.ipynb",
         "append_to_reM": "14_synchro.nested_stims.ipynb",
         "master_loop": "14_synchro.nested_stims.ipynb",
         "load_vivo_2p": "99_testdata.ipynb"}

modules = ["core.py",
           "utils.py",
           "processing.py",
           "modelling.py",
           "plotting.py",
           "database.py",
           "eyetrack.py",
           "synchro/io.py",
           "synchro/extracting.py",
           "synchro/processing.py",
           "leddome.py",
           "synchro/nested_stims.py",
           "testdata.py"]

doc_url = "https://Tom-TBT.github.io/theonerig/"

git_url = "https://github.com/Tom-TBT/theonerig/tree/master/"

def custom_doc_links(name): return None
